# from .loggers import create_new_logger, module_logger
from .pretty_printer import printify

# Import formatters
from .formatters import ColoredFormatter
from .formatters_celery import CeleryTaskFormatter
from .formatters_celery import CeleryTaskFormatterColored



import logging
import sys
from pathlib import Path

ceon_render_dev_dir = (
    "/mnt/FileStorage/Dayne/Web/proj_ceonmedia/assets/python/ceon_render"
)
sys.path.append(ceon_render_dev_dir)


from ceonmedia import json_io

# from ceonmedia import CstockProjInput, CstockRenderTask
# from ceonmedia import CstockProject
# from ceonmedia.core.project import CstockProjectInfo
from ceonmedia.log import formatters, printify

# from ceonmedia.filesystem import CstockLocalProject
from ceonmedia import json_io

# Logger setup
root_logger = logging.getLogger()
sh = logging.StreamHandler()
sh.setFormatter(formatters.DEV)
root_logger.addHandler(sh)
root_logger.setLevel("DEBUG")

logger = logging.getLogger(__name__)

PROJECTS_DIR = (
    "/mnt/FileStorage/Dayne/Web/proj_ceonmedia/assets/my_cstock_projects"
)
PROJECT_FLDR = "Waving_Flag"
PROJECT_PATH = Path(PROJECTS_DIR, PROJECT_FLDR)


def test_io_project():
    """Load a project and then export it."""
    file_to_load = "/mnt/FileStorage/Dayne/Web/proj_ceonmedia/assets/python/ceonmedia/tests/test_json_io/ceonmedia_project.json"
    if not Path(file_to_load).exists():
        raise FileNotFoundError(f"File not found: {file_to_load}")
    loaded_project = json_io.project.from_file(file_to_load)
    print(loaded_project)

    print("Saving...")
    out_file = Path(file_to_load).with_name("testprojectoutput.json")
    json_io.project.to_file(str(out_file), loaded_project)
    print(f"Saved: {out_file}")


def test_io_job():
    """Load a job and then export it."""
    file_to_load = "/mnt/FileStorage/Dayne/Web/proj_ceonmedia/assets/python/ceonmedia/tests/test_json_io/ceonmedia_job.json"
    if not Path(file_to_load).exists():
        raise FileNotFoundError(f"File not found: {file_to_load}")
    loaded_project = json_io.job.from_file(file_to_load)
    print(loaded_project)

    print("Saving...")
    out_file = Path(file_to_load).with_name("testjoboutput.json")
    json_io.job.to_file(str(out_file), loaded_project)
    print(f"Saved: {out_file}")


def old_stuff():
    # Proj Inputs
    # proj_input1 = CstockProjInput(
    #     name="my_proj_input_1",
    #     job_input_type=CstockProjInput.job_input_types.IMG,
    #     web_input_type=CstockProjInput.web_input_types.IMG,
    # )

    # local_project = CstockLocalProject(f"{PROJECT_PATH}")
    # print()
    # logger.info("Got local project: ")
    # logger.info(printify(local_project))
    # print()

    # json_data = local_project.json()
    # print()
    # logger.info("Got local project json: ")
    # logger.info(printify(json_data))
    # print()

    # cstock_project = local_project.load()
    # print()
    # logger.info("Got loaded_data: ")
    # logger.info(cstock_project)
    # print()

    # # Test export
    # local_project_export_file = "test_project_export.json"
    # print()
    # as_json = json_io.exporter.dumps(cstock_project)
    # logger.info(f"As json:\n{as_json}")
    # with open("local_project_export_file.json", "w+") as f:
    #     f.write(as_json)
    # logger.warning(f"Writtent to file: {local_project_export_file}")

    # Render Tasks
    # render_task1 = CstockRenderTask(
    #     name="my_render_task",
    #     app_type=CstockRenderTask.app_types.HOU,
    #     app_version="19.5",
    #     target_file="relpath/to/file.hiplc",
    #     output="render_output.$F4.exr",
    #     app_render_settings={},
    # )
    # render_tasks = [render_task1, render_task2]

    # test_io_project()
    pass


def main():
    test_io_job()
    test_io_project()


if __name__ == "__main__":
    main()



import json
import inspect
import logging
from pathlib import Path
from typing import Union
from typing import TypeVar
from typing import Type
from ceonmedia import errors

logger = logging.getLogger()

T = TypeVar("T")


def loads(json_string: str):
    return json.loads(json_string)


def load(json_filepath: Union[Path, str]):
    if not Path(json_filepath).is_file():
        raise FileNotFoundError(
            f"Unable to find json file for json_file_path: {json_filepath}"
        )
    with open(str(json_filepath)) as f:
        json_data = json.load(f)
    return json_data


def cstock_from_dict(dict_kwargs: dict, cls_to_instantiate: Type[T]) -> T:
    """
    Init from a dict, but fail and reporting if unrecognized kwargs are found.
    -
    This was added because instantiating with cls(**kwargs_dict) silently ignores unrecognized kwargs
    when running in FastAPI (FastAPI overrides dataclasses default behaviour with pydantic versions???).
    This meant that projects which were identified as valid in the webUI/FileServer would fail when trying to run
    on the render server (if unrecognized kwargs were found).
    Therefore, use this function to instantiate classes from a dict.
    """
    unrecognized_kwargs = []
    known_kwargs = {}
    for k, v in dict_kwargs.items():
        if k in inspect.signature(cls_to_instantiate).parameters:
            known_kwargs[k] = v
        else:
            unrecognized_kwargs.append(k)

    if unrecognized_kwargs:
        logger.warning(
            f"Received {len(unrecognized_kwargs)} unrecognized kwargs: {unrecognized_kwargs}"
        )
        raise errors.CstockInstantiationError(
            cstock_cls=cls_to_instantiate,
            message=f"Unrecognized kwargs: {unrecognized_kwargs}",
        )
    try:
        return cls_to_instantiate(**known_kwargs)
    except TypeError as e:
        # Catch TypeError if cls_to_instantiate is missing required args. Re-raise as InstantiationError
        raise errors.CstockInstantiationError(cstock_cls=cls_to_instantiate, errors=e,message="Failed to instantiate class instance from dict.")


def cstock_from_json(
    json_filepath: Union[Path, str], cstock_cls: Type[T]
) -> T:
    json_data = load(json_filepath)
    cstock_instance = cstock_from_dict(json_data, cstock_cls)
    return cstock_instance



from ceonmedia.core.file_action import CstockFileActionResize
from ceonmedia.core.file_action import CstockFileActionType

# Examples that should always work
SUCCESS_CASES = [
    # TODO missing OPTIONAL kwarg.
    {
        # As raw strings/dict
        "action_type": "crop",
        "action_kwargs": {"x": 20, "y": 20, "width": 30, "height": 40},
    },
    {
        # As raw strings/dict
        "action_type": "resize",
        "action_kwargs": {"max_width": 1920, "max_height": 1080},
    },
    {
        # with enum instance as type
        "action_type": CstockFileActionType.RESIZE,
        "action_kwargs": {"max_width": 1920, "max_height": 1080},
    },
    {
        # With class instance nested in dict
        "action_type": "resize",
        "action_kwargs": CstockFileActionResize(
            max_width=1920, max_height=1080
        ),
    },
    {
        # With enum and class instance as args.
        "action_type": CstockFileActionType.RESIZE,
        "action_kwargs": CstockFileActionResize(
            max_width=1920, max_height=1080
        ),
    },
    {
        # With class instance nested in dict
        "action_type": "resize",
        "action_kwargs": CstockFileActionResize(
            max_width=1920, max_height=1080
        ),
    },
]

# Examples that we know should not work.
FAIL_CASES = [
    {
        # Invalid parm name
        "action_type": "resize",
        "action_kwargs": {"typo_kwargname": 1920, "max_height": 1080},
    },
    {
        # Args are valid for CROP type, but not for this action_type.
        "action_type": "resize",
        "action_kwargs": {"x": 20, "y": 20, "width": 30, "height": 40},
    },
    {
        # Missing required crop kwarg 'y'
        "action_type": "crop",
        "action_kwargs": {"x": 20, "width": 30, "height": 40},
    },
]



# DEPRECATED: Replace with ceon-render module.

import logging
from dataclasses import dataclass, field
from typing import ClassVar, Type
from typing import Optional
from abc import ABC
from uuid import UUID, uuid4

from .base import CstockBaseClass, CstockBaseEnum
from .file_reference import CstockFileReference
from . import validator
from ceonmedia import errors

logger = logging.getLogger(__name__)


class CstockRenderAppType(CstockBaseEnum):
    HOU = "hou"
    FFMPEG = "ffmpeg"
    NUKE = "nuke"


@dataclass(kw_only=True)
class CstockRenderTaskAppSettings(CstockBaseClass, ABC):
    pass


@dataclass(kw_only=True)
class CstockRenderTask(CstockBaseClass):
    # So that the user can access types without explicitly needing to import CstockRenderAppType
    app_types: ClassVar[Type[CstockRenderAppType]] = CstockRenderAppType

    task_name: str
    app_type: CstockRenderAppType
    app_version: str
    # app_render_settings will be different depending on the app type
    # (e.g. Hou needs a path to a node, FFMPEG takes strings for input/outputs args)
    app_render_settings: CstockRenderTaskAppSettings

    task_input: CstockFileReference  # Input file
    task_output: str  # Output file

    # Store other render tasks which must be completed before this one can start
    task_dependencies: list[str] = field(default_factory=list)
    task_uuid: UUID = field(default_factory=uuid4)

    def __post_init__(self):
        self.app_type = validator.enforce_enum(
            self.app_type, CstockRenderAppType
        )
        self.task_uuid = validator.enforce_uuid(self.task_uuid)
        self.task_input = validator.enforce_class_instance(
            self.task_input, CstockFileReference
        )

        # Handle dynamic loading of AppSettings class instances
        cls_to_load = APP_SETTINGS_CLS[self.app_type]
        if not isinstance(self.app_render_settings, cls_to_load):
            logger.debug(
                f"Loading app_settings class '{cls_to_load}' from data: {self.app_render_settings}"
            )
            self.app_render_settings = cls_to_load(**self.app_render_settings)
        # except (ValueError, TypeError) as e:
        #     raise errors.CstockInstantiationError(
        #         cstock_cls=self.__class__, errors=[e]
        #     )
        # self.validate()

    # def validate(self):
    #     # Ensure that the correct RenderSettings were provided for the corresponding app type.
    #     expected_cls = APP_SETTINGS_CLS[self.app_type]

    def __str__(self):
        msg = f"<{self.__class__.__name__} '{self.task_name}' ({self.app_type} OUT:'{self.task_output}')>"
        return msg

    def __repr__(self):
        return self.__str__()


@dataclass(kw_only=True)
class CstockRenderTaskAppSettingsFFMPEG(CstockRenderTaskAppSettings):
    app_type: ClassVar[CstockRenderAppType] = CstockRenderAppType.FFMPEG
    input_args: str = ""
    output_args: str = ""


@dataclass(kw_only=True)
class CstockRenderTaskAppSettingsHou(CstockRenderTaskAppSettings):
    app_type: ClassVar[CstockRenderAppType] = CstockRenderAppType.HOU
    target_node: str
    frames: str
    out_dimensions: Optional[str] = None


APP_SETTINGS_CLS = {
    CstockRenderAppType.HOU: CstockRenderTaskAppSettingsHou,
    CstockRenderAppType.FFMPEG: CstockRenderTaskAppSettingsFFMPEG,
}



import logging
import inspect
from typing import List, Optional, Type
from dataclasses import dataclass, field
from uuid import UUID, uuid4

from ceon_render import CeonRenderPipeline

from .base import CstockBaseClass

# from .render import CstockRenderPipeline
from .project_input import CstockProjectInput
from . import validator

# from .render import CstockRenderPipeline

logger = logging.getLogger(__name__)


@dataclass(kw_only=True)
class CstockProjectInfo(CstockBaseClass):
    """
    Information about this project that is only needed for the website.
    """

    title: str
    description: str = ""
    tags: Optional[List[str]] = field(default_factory=list)


@dataclass(kw_only=True)
class CstockProject(CstockBaseClass):
    """Fundamental data for a Cstock project to function."""

    project_inputs: List[CstockProjectInput]
    render_pipeline: CeonRenderPipeline
    project_info: Optional[CstockProjectInfo] = None
    project_id: str = field(default_factory=lambda: str(uuid4()))

    def __post_init__(self):
        self.project_inputs = validator.enforce_list_of_instances(
            self.project_inputs, CstockProjectInput
        )
        if self.project_info:
            self.project_info = validator.enforce_class_instance(
                self.project_info, CstockProjectInfo
            )
        self.project_id = str(self.project_id)
        self.render_pipeline = validator.enforce_class_instance(
            self.render_pipeline, CeonRenderPipeline
        )

    def project_input(self, project_input_name: str) -> CstockProjectInput:
        """Fetch a project input by name"""
        target_input = None
        for project_input in self.project_inputs:
            if project_input.name == project_input_name:
                target_input = project_input
                break
        if not target_input:
            raise Exception(
                f"Could not find target project_input: {project_input_name}"
            )
        return target_input

    def validate(self):
        if not isinstance(self.project_inputs, list):
            raise Exception(
                f"CstockProject project_inputs is not a list: {self.project_inputs}"
            )
        for project_input in self.project_inputs:
            if not isinstance(project_input, CstockProjectInput):
                raise Exception(
                    f"CstockProject contains a project_input which is not an instantiated cstock instance: {project_input}"
                )
        if not isinstance(self.render_pipeline, CeonRenderPipeline):
            raise Exception(
                f"CstockProject contains a render_pipeline which is not a cstock instance: {self.render_pipeline}"
            )
        if not isinstance(self.project_info, CstockProjectInfo):
            raise Exception(
                f"CstockProject contains project_info which is not a cstock instance: {self.project_info}"
            )
        if not isinstance(self.project_id, str):
            raise Exception(
                f"CstockProject project_id is not a str instance: {self.project_id}"
            )



from dataclasses import dataclass, field
from typing import Callable, ClassVar, TypeVar
from typing import Union
from typing import Dict
from abc import ABC
from ceonmedia.core.base import CstockBaseEnum
from ceonmedia.core.base import CstockBaseClass


class CstockFileActionType(CstockBaseEnum):
    DOC_TO_IMGS = "doc_to_imgs"
    RESIZE = "resize"
    CROP = "crop"


# All file action kwargs inherit from this base class.
# These are used to validate the file action kwargs according to the type.
class CstockFileActionKwargs(ABC):
    pass


# The CstockFileAction class is to be used for storing file actions, converting to/from json.
@dataclass
class CstockFileAction(CstockBaseClass):
    action_type: Union[str, CstockFileActionType]
    action_kwargs: Union[Dict, CstockFileActionKwargs]

    def __str__(self):
        return f"<{self.__class__.__name__} {self.action_type}: {self.action_kwargs}>"

    def __post_init__(self):
        """Validate kwargs"""
        # Enforce enum type
        if not isinstance(self.action_type, CstockFileActionType):
            self.action_type = CstockFileActionType(self.action_type)
        # Enforce kwargs type
        loadable_classes = [
            CstockFileActionCrop,
            CstockFileActionResize,
            CstockFileActionDocToImgs,
        ]
        kwargs_lookup = {
            file_action.action_type: file_action
            for file_action in loadable_classes
        }
        if not isinstance(self.action_kwargs, CstockFileActionKwargs):
            cls_to_load = kwargs_lookup[self.action_type]
            loaded_cls = cls_to_load(**self.action_kwargs)
            self.action_kwargs = loaded_cls


@dataclass
class CstockFileActionCrop(CstockFileActionKwargs):
    action_type: ClassVar[CstockFileActionType] = CstockFileActionType.CROP
    # Crop settings handled as PERCENTAGES.
    x: float
    y: float
    width: float
    height: float


@dataclass
class CstockFileActionResize(CstockFileActionKwargs):
    action_type: ClassVar[CstockFileActionType] = CstockFileActionType.RESIZE
    max_width: int
    max_height: int


# For types that don't accept any arguments
@dataclass
class CstockFileActionDocToImgs(CstockFileActionKwargs):
    action_type: ClassVar[
        CstockFileActionType
    ] = CstockFileActionType.DOC_TO_IMGS


# def get_kwargs(
#     file_action_type: CstockFileActionType,
# ) -> CstockFileActionKwargs:



import json
from pathlib import Path

import core.job_input as cjob_input
import core.proj_input as cproj_input


def load_job_inputs(job_inputs: list[dict]) -> list[cjob_input.CstockJobInput]:
    loaded_job_inputs = list(map(cjob_input.CstockJobInput.from_dict, job_inputs))
    return loaded_job_inputs


def load_proj_inputs(proj_inputs: list[dict]) -> list[cproj_input.CstockProjInput]:
    loaded_proj_inputs = list(map(cproj_input.CstockProjInput.from_dict, proj_inputs))
    return loaded_proj_inputs



# Pair special keys with specific loaders
LOADERS = {
    "job_inputs": load_job_inputs,
    "proj_inputs": load_proj_inputs
}


def load(json_filepath: str, raise_empty_file=True):
    # read json file
    if not Path(json_filepath).is_file():
        raise FileNotFoundError(
            f"Unable to find json file for json_file_path: {json_filepath}"
        )
    json_data = {}
    # Get data from file
    try:
        with open(str(json_filepath)) as f:
            json_data = json.load(f)
    except json.decoder.JSONDecodeError as e:
        raise Exception(f"Json failed to decode file for {json_filepath}: {e}")

    if not len(json_data) > 0:  # No data in file
        # print("No data found in file")
        if raise_empty_file:
            raise Exception(f"File does not contain data: {json_filepath}")
    # print(f"Got data in JSON({json_filepath}): {json_data}")

    loaded_data = dict_to_cstock(json_data)
    return loaded_data


def dict_to_cstock(json_dict_data: dict) -> dict:
    new_dict = {}
    for key in json_dict_data.keys():
        target_data = json_dict_data[key]
        if key in LOADERS.keys():
            loader_fn = LOADERS[key]
        else:
            loader_fn = None
        # print(f"key: {key}, loader: {loader_fn}")
        loaded_objects = loader_fn(target_data)
        # print(f"loaded objects: ")
        # for loaded_object in loaded_objects:
            # print(f"\t{loaded_object}")
        new_dict[key] = loaded_objects
    return new_dict





from dataclasses import dataclass
from typing import Optional
from abc import ABC
# from dataclasses import dataclass
from enum import Enum

from typing import Optional, Type, Dict, TypeVar
from ceonmedia.log import create_new_logger
from .. import errors

logger = create_new_logger(__name__)

# P = TypeVar("P", bound="ICstockRenderAppSettings")

class CstockRenderAppType(Enum):
    HOU = "hou"
    FFMPEG = "ffmpeg"
    NUKE = "nuke"

    # def get_class(self):
    # return get_app_settings_cls(self)

    # def make_app_settings(self, kwargs=None):
    # return make_app_settings(self, kwargs)


class RenderEngineType(Enum):
    MANTRA = "mantra"  # Houdini old engine
    KARMA = "karma"  # Houdini new engine
    REDSHIFT = "redshift"  # Redshift


# For type hinting/ide to recognize inherited classes
class ICstockRenderAppSettings(ABC):
    pass

# -- App Settings classes--
# Contains settings which are required to execute a render in the target app.
# --


@dataclass
class CstockRenderAppSettingsFFmpeg(ICstockRenderAppSettings):
    input_args: str
    output_args: str


@dataclass
class CstockRenderAppSettingsHou(ICstockRenderAppSettings):
    target_node: str
    nodetype: str
    frames: str
    render_engine: Optional[RenderEngineType] = None
    take: str = "Main"


def get_app_settings_cls(
    app: CstockRenderAppType,
) -> Type[ICstockRenderAppSettings]:
    CLASS_LOOKUP = {
        CstockRenderAppType.HOU: CstockRenderAppSettingsHou,
        CstockRenderAppType.FFMPEG: CstockRenderAppSettingsFFmpeg,
    }
    try:
        type_class = CLASS_LOOKUP[app]
    except KeyError:
        raise errors.UnknownRenderAppTypeError(app)
    return type_class


def make_app_settings(
    app: CstockRenderAppType, kwargs: Optional[Dict] = None
) -> ICstockRenderAppSettings:
    if not kwargs:
        kwargs = {}
    app_settings_cls = get_app_settings_cls(app)
    logger.debug(f"Creating {app_settings_cls} with kwargs: ", kwargs)
    app_settings = app_settings_cls(**kwargs)
    logger.debug(f"Created instance: {app_settings}")
    return app_settings



# A helper class that loads a target dir as a CstockProj
# Useful for validating the integrity of the job files
# Useful for identifying the file path that represents the job in a server with a filesystem
from __future__ import annotations
import logging
from typing import Union, Optional, List, Literal
from dataclasses import dataclass
from pathlib import Path

from ceonmedia import json_io
from ceonmedia import CstockJob

# from ceonmedia.core.base import T
logger = logging.getLogger(__name__)


JOB_FOLDERS: dict[Literal["job_inputs", "job_outputs", "logs"], str] = {
    "job_inputs": "job_inputs",
    "job_outputs": "job_outputs",
    "logs": "logs",
}


class CstockLocalJobDirs:
    def __init__(self, job_root_dir: Union[Path, str]):
        self.root = Path(job_root_dir)
        self.job_inputs = Path(job_root_dir, "job_inputs")
        self.job_outputs = Path(job_root_dir, "job_outputs")
        self.logs = Path(job_root_dir, "logs")

    def create_directories(self):
        paths_to_make = []
        for path in self.__dict__.values():
            if not Path(path).exists():
                paths_to_make.append(path)
        stringified_paths = "\n\t".join([str(path) for path in paths_to_make])
        logger.info(f"Creating new folder(s):\n\t{stringified_paths}")
        for path in paths_to_make:
            Path(path).mkdir(parents=True, exist_ok=True)


class CstockLocalJob:
    json_file_name = "ceonmedia_job.json"
    """
    Represents a job in the local filesystem.
    Provides functionality for loading data from files.
    """

    def __init__(
        self,
        job_dir_path: Union[Path, str],
        json_file_name: Optional[str] = None,
    ):
        self._dirs = CstockLocalJobDirs(job_dir_path)
        json_file_name = (
            json_file_name if json_file_name else CstockLocalJob.json_file_name
        )
        self._json_file_path = Path(self._dirs.root, json_file_name)
        if not self._dirs.root.is_dir():
            raise NotADirectoryError(
                f"LocalJob dir does not exist: {self._dirs.root}"
            )
        if not self._json_file_path.is_file():
            raise FileNotFoundError(
                f"LocalJob {self._dirs.root} missing expected JSON file: {self._json_file_path.name}"
            )

    def path(self) -> Path:
        return self._dirs.root

    def json_path(self) -> Path:
        return self._json_file_path

    def dirs(self) -> CstockLocalJobDirs:
        return CstockLocalJobDirs(self.path())

    def load(self) -> CstockJob:
        """
        Loads a CstockJob instance from the json file.
        Loading imports all data as python/ceonmedia classes
        """
        logger.debug(f"Loading CstockJob({self.path()})...")
        cstock_job = json_io.job.from_file(
            str(self._json_file_path.absolute())
        )
        return cstock_job

    def __str__(self):
        return f"<{self.__class__.__name__} {self.path()}>"

    def __repr__(self):
        return self.__str__()


def create_local_job(
    job_dir: Union[str, Path], ceonmedia_job: CstockJob
) -> CstockLocalJob:
    """Sets up the folders in the local filesystem and creates ceonmedia_job.json"""
    if Path(job_dir).exists():
        raise Exception(
            f"Cannot create local job because the directory already exists: {job_dir}"
        )

    logger.info(f"{job_dir=}, {ceonmedia_job=}")
    CstockLocalJobDirs(job_dir).create_directories()
    json_path = str(Path(job_dir, CstockLocalJob.json_file_name))
    json_io.job.to_file(json_path, ceonmedia_job)

    ceonmedia_local_job = CstockLocalJob(job_dir)
    return ceonmedia_local_job



import logging
import json
from dataclasses import asdict
from typing import List, Dict, Any, Callable
from typing import Optional, Union, Type
from typing import Iterable

logger = logging.getLogger(__name__)

BRACES_LIST = ("[", "]")
BRACES_DICT = ("{", "}")


def printify(data: Any, indent: int = 1) -> str:
    logger.debug(f"Got data of type {type(data)}: {data}")
    """ Returns data object as a string formmatted for printing """
    fn = get_printify_function(data)
    logger.debug(f"Chosen fn: {fn}")
    as_str = fn(data, indent=indent)
    return as_str


def get_printify_function(data: Iterable) -> Callable:
    LOOKUP = {dict: printify_dict, list: printify_list}
    logger.debug(f"LOOKUP: {LOOKUP}")
    try:
        fn = LOOKUP[type(data)]
    except KeyError as e:
        logger.debug(f"KeyError: {e}")
        try:
            # Check if the object is a class instance containing a __dict__ method.
            test = data.__dict__
            return printify_class
        except AttributeError:
            logger.warn(f"Invalid data type: {type(data)}")
        return printify_str
    logger.debug(f"fn: {fn}")
    return fn


def printify_str(data, indent=1) -> str:
    tabs = "\t" * indent
    return tabs + str(data)


# def printify_dict(data: Dict, indent: int = 1) -> str:
#     """Returns data object as a string formmatted for printing"""
#     dict_to_print = {key: str(value) for (key, value) in data.items()}
#     tabs = "\t" * indent
#     lines = []
#     for key, value in dict_to_print.items():
#         lines.append(f"{tabs}{key}:\t{value}")
#     _add_surrounding_braces(lines, BRACES_DICT, indent=indent - 1)
#     as_str = "\n".join([line for line in lines])
#     return as_str


def printify_dict(data: Dict, indent: int = 1) -> str:
    """Returns data object as a string formmatted for printing"""
    dict_to_print = {key: str(value) for (key, value) in data.items()}
    return json.dumps(dict_to_print, indent=4)


def printify_list(data: list, indent: int = 1) -> str:
    """Returns list object as a string formmatted for printing"""
    list_to_print = [str(value) for value in data]
    tabs = "\t" * indent

    lines = []
    for item in list_to_print:
        lines.append(f"{tabs}{item}")
    _add_surrounding_braces(lines, indent=indent - 1, brace_chars=BRACES_LIST)
    as_str = "\n".join([line for line in lines])
    return as_str


def printify_class(data, indent: int = 1) -> str:
    # Printify a class by converting it to a dict.
    try:
        return printify_dict(asdict(data))
    except Exception as e:
        logger.warning(f"TODO: handle not-a-dataclass error: {e}")
    return printify_dict(data.__dict__, indent=indent)


def _add_surrounding_braces(
    lines: list[str], brace_chars: tuple[str, str], indent=0
):
    """Add braces to the start and end of the list containing the lines to be printed."""
    indent = max(0, indent)
    brace_tabs = "\t" * indent
    lines.insert(0, f"{brace_tabs}{brace_chars[0]}")
    lines.append(f"{brace_tabs}{brace_chars[1]}")



from ceonmedia.core.project_input import CstockProjInput

# from ceonmedia.core.job_input import CstockJobInputType
# from ceonmedia.core.web_input import CstockWebInputType

from . import json_tests
from ceonmedia.tests.dicts import proj_input_dicts

TEST_CLS = CstockProjInput
TEST_DICTS = proj_input_dicts


def test_proj_input_json_serialization() -> None:
    for test_dict in TEST_DICTS.SUCCESS_CASES:
        test_object = TEST_CLS(**test_dict)
        json_tests.run_json_serialization_tests(cstock_object=test_object)



import logging
from ceonmedia import CstockProject
from ceonmedia import CstockProjectInput
from ceonmedia import CstockJob
from ceonmedia import CstockJobInputType
from ceonmedia import CstockWebInputType
from ceonmedia.core.project import CstockProjectInfo
from ceonmedia.core.project import CstockRenderPipeline
from ceonmedia.log import printify

from ceonmedia.job_maker import project_preview_video_maker
from ceonmedia.job_maker import job_maker

from ceonmedia.log import formatters, formats

root_logger = logging.getLogger()
root_logger.setLevel("INFO")

formatter = formatters.ColoredFormatter(formats.DEV)
sh = logging.StreamHandler()
sh.setFormatter(formatter)

root_logger.addHandler(sh)

logger = logging.getLogger(__name__)

PROJECT_INPUT_PREVIEW_FILES_DIR_DEFAULT = "/mnt/FileStorage/Dayne/Web/proj_ceonmedia/assets/project_input_files/project_preview_video_inputs/default_inputs"
PROJECT_INPUT_PREVIEW_FILES_DIR_EXAMPLES = "/mnt/FileStorage/Dayne/Web/proj_ceonmedia/assets/project_input_files/project_preview_video_inputs/example_inputs"
JOBS_DIR = "/mnt/FileStorage/Dayne/Web/proj_ceonmedia/local_storage/jobs"

# PROJECT_INPUTS = [
#     CstockProjectInput(
#         name="pen_color",
#         job_input_type=CstockJobInputType.COLOR,
#         web_input_type=CstockWebInputType.COLOR,
#         description="Color of pen",
#     ),
#     CstockProjectInput(
#         name="flag_image",
#         job_input_type=CstockJobInputType.IMG,
#         web_input_type=CstockWebInputType.IMG,
#         description="Image on flag",
#     ),
#     CstockProjectInput(
#         name="box_color",
#         job_input_type=CstockJobInputType.COLOR,
#         web_input_type=CstockWebInputType.COLOR,
#         description="Color of box",
#     ),
# ]
PROJECT_INPUTS = [
    CstockProjectInput(
        name="single_image",
        job_input_type=CstockJobInputType.IMG,
        web_input_type=CstockProjectInput.web_input_types.IMG,
        description="A single image TODO system to enforce limit on num inputs",
    ),
    CstockProjectInput(
        name="image_from_doc",
        job_input_type=CstockJobInputType.IMG,
        web_input_type=CstockProjectInput.web_input_types.DOC,
        description="A single image TODO system to enforce limit on num inputs",
    ),
    CstockProjectInput(
        name="single_color",
        job_input_type=CstockJobInputType.COLOR,
        web_input_type=CstockProjectInput.web_input_types.COLOR,
        description="A single color",
    ),
    CstockProjectInput(
        name="dropdown_string",
        job_input_type=CstockJobInputType.STRING,
        web_input_type=CstockProjectInput.web_input_types.DROPDOWN,
        type_settings={
            "dropdown_options": [
                {
                    "value": "dropdown_option_value_0",
                    "label": "Option Label 0",
                },
                {
                    "value": "dropdown_option_value_1",
                    "label": "Option Label 1",
                },
                {
                    "value": "dropdown_option_value_2",
                    "label": "Option Label 2",
                },
            ]
        },
        description="A single image TODO system to enforce limit on num inputs",
    ),
]

RENDER_PIPELINE = CstockRenderPipeline(
    render_tasks=[], primary_output="out", extra_outputs=[]
)

PROJECT_INFO = CstockProjectInfo(
    "My Test Project",
    description="Testing the genration of jobs for hte preview video",
    tags=[],
)

PROJECT_TO_TEST = CstockProject(
    project_inputs=PROJECT_INPUTS,
    render_pipeline=RENDER_PIPELINE,
    project_info=PROJECT_INFO,
)


def main(ceonmedia_project: CstockProject):
    project_input_jobs: dict[
        str, list[CstockJob]
    ] = project_preview_video_maker.create_jobs_for_project_preview_video(
        ceonmedia_project,
        files_dir_default=PROJECT_INPUT_PREVIEW_FILES_DIR_DEFAULT,
        files_dir_examples=PROJECT_INPUT_PREVIEW_FILES_DIR_EXAMPLES,
    )
    print_header("Created project_input jobs:")
    logger.info(f"{printify(project_input_jobs)}")

    # Test execute
    print()
    project_input_name_to_test: str = list(project_input_jobs.keys())[1]
    print_header(
        f"Testing with isolated project input: {project_input_name_to_test}"
    )
    jobs_for_project_input = project_input_jobs[project_input_name_to_test]
    file_sources = [
        PROJECT_INPUT_PREVIEW_FILES_DIR_DEFAULT,
        PROJECT_INPUT_PREVIEW_FILES_DIR_EXAMPLES,
    ]
    for job in jobs_for_project_input:
        logger.info(f"{job}")
        job_maker.create_cstock_job_on_disk(job, JOBS_DIR, file_sources)


def print_header(msg):
    logger.info(f"*****\n{msg}\n*****")


if __name__ == "__main__":
    main(PROJECT_TO_TEST)



import logging
from ceonmedia.log import formats
from enum import Enum


class TerminalColor(Enum):
    GREY = "\x1b[38;20m"
    GREEN = "\x1b[32m"
    BLUE = "\x1b[34m"
    CYAN = "\x1b[36m"
    YELLOW = "\x1b[33;20m"
    RED = "\x1b[31;20m"
    RED_BOLD = "\x1b[31;1m"
    _RESET = "\x1b[0m"



import logging
from uuid import uuid4
from pathlib import Path
from typing import Optional, Union
from shutil import copy
from uuid import UUID

from ceonmedia import CstockJob
from ceonmedia.filesystem import local_job

logger = logging.getLogger(__name__)


def create_cstock_job_on_disk(
    cstock_job: CstockJob,
    jobs_dir: Union[str, Path],
    file_source_dirs: Union[list[Path], list[str]],
) -> UUID:
    """
    Create a job directory with ceonmedia_job.json file and job_inputs files.
    jobs_dir: The root directory that contains all jobs.
    file_source_dirs: The dirs in where to look for input files.
    Returns: The uuid of the newly created job
    """
    logger.info(f"{cstock_job=} {jobs_dir=}, {file_source_dirs=}")
    job_uuid = cstock_job.job_uuid
    job_dir = Path(jobs_dir, str(job_uuid))
    # Creates the fodlers and json file
    cstock_local_job = local_job.create_local_job(
        job_dir=job_dir, ceonmedia_job=cstock_job
    )

    # Copy the input files from value_picker source to job_inputs dir.
    for job_input in cstock_job.job_inputs:
        if not job_input.is_file_type:
            continue

        for value in job_input.values:
            source_file_path = _get_file_path(file_source_dirs, value)
            dest_path = Path(cstock_local_job.dirs().job_inputs, value)
            logger.info(
                f"Copying file:\n\tFrom: {source_file_path}\n\tTo: {dest_path}"
            )
            copy(source_file_path, dest_path)
    return UUID(job_uuid)


def _get_file_path(
    dirs_to_check: Union[list[str], list[Path]], file_name: str
) -> Path:
    """Receive a file_name and return the path to the source file"""
    dirs_to_check = [str(path) for path in dirs_to_check]
    files = []
    for dirpath in dirs_to_check:
        files += _list_files_in_dir(dirpath)
    for file in files:
        if Path(file).name == file_name:
            return file
    raise Exception(f"Could not find source file from name: {file_name}")


def _list_files_in_dir(dir: Union[Path, str]) -> list[Path]:
    file_dir = Path(dir)
    if not file_dir.exists():
        raise NotADirectoryError(file_dir)
    filelist = file_dir.glob("**/*")
    files = [file for file in filelist if file.is_file()]
    logger.debug(f"Got {len(files)} files:")
    for file in files:
        logger.debug(f"\t{file}")
    return files



from pathlib import Path

from ceonmedia.log import logging_info, module_logger, logging_setup, printify
from ceonmedia import json_io
from ceonmedia import CstockJobInput, CstockProjInput, CstockRenderTask

# Logger setup
logging_setup.setup_root_logger()

logger = module_logger()


if __name__ == "__main__":
    # Testing github changes
    filename = "tempTests/myTestJSON.json"
    print(f"Exporting to: {Path(filename)}")

    # Job Inputs
    job_input1 = CstockJobInput(
        "my_job_input_1", CstockJobInput.types.IMG, values=["relpath/to/somefile.jpg"]
    )
    job_input2 = CstockJobInput(
        "my_job_input_2",
        CstockJobInput.types.IMG,
        values=["relpath/to/someotherfile.jpg"],
    )
    job_inputs = [job_input1, job_input2]

    # Proj Inputs
    proj_input1 = CstockProjInput(
        name="my_proj_input_1",
        job_input_type=CstockProjInput.job_input_types.IMG,
        web_input_type=CstockProjInput.web_input_types.IMG,
    )
    proj_input2 = CstockProjInput(
        name="my_proj_input_2",
        job_input_type=CstockProjInput.job_input_types.IMG,
        web_input_type=CstockProjInput.web_input_types.IMG,
    )
    proj_inputs = [proj_input1, proj_input2]

    # Render Tasks
    render_task1 = CstockRenderTask(
        name="my_render_task",
        app_type=CstockRenderTask.app_types.HOU,
        app_version="19.5",
        target_file="relpath/to/file.hiplc",
        output="render_output.$F4.exr",
        app_render_settings={},
    )
    render_task2 = CstockRenderTask(
        name="my_render_task2",
        app_type=CstockRenderTask.app_types.HOU,
        app_version="19.5",
        target_file="relpath/to/file.hiplc",
        output="render_output.$F4.exr",
        app_render_settings={},  # TODO target node, render engine etc
    )
    render_tasks = [render_task1, render_task2]

    # Export
    dict_data = {
        "proj_inputs": proj_inputs,
        "render_tasks": render_tasks,
        "job_inputs": job_inputs,
    }
    json_io.write_json_file(dict_data, filename)

    print(f"Loading from file: {filename}")
    loaded_data = json_io.read_json_file(filename)
    print(printify(loaded_data))



from dataclasses import dataclass, field
from typing import List, Type, ClassVar
from typing import Optional
from enum import Enum
from uuid import UUID, uuid4

from . import validator
from .job_input import CstockJobInput
from .base import CstockBaseEnum
from .base import CstockBaseClass


class CstockJobType(CstockBaseEnum):
    PREVIEW = "video_preview"  # Deliver a low resolution watermarked sample video.
    VIDEO_HD = "video_hd"  # Deliver a 1920x1080p video.
    VIDEO_4K = "video_4k"  # Deliver a 3840x2160 video.
    # VIDEO_CUSTOM = "video_custom" # future expansion to allow render args to be passed

    # def get_job_settings(self):
    # return get_job_settings(self)


@dataclass
class CstockJob(CstockBaseClass):
    # For easy developer access without needing explicitly import CstockJobType
    job_types: ClassVar[Type[CstockJobType]] = CstockJobType

    job_inputs: List[CstockJobInput]
    project_uuid: str  # Can be looked up if not provided?
    job_type: CstockJobType = CstockJobType.PREVIEW
    job_uuid: str = field(default_factory=lambda: str(uuid4()))
    limit_frames: Optional[str] = None
    # job_settings: args for cusom setups

    def job_input(self, job_input_name: str) -> CstockJobInput:
        """Lookup a job input by name"""
        for job_input in self.job_inputs:
            if job_input.name == job_input_name:
                return job_input
        raise ValueError(f"No job input found for name '{job_input_name}'")

    def __post_init__(self):
        # TODO use 'validator' to enforce enum for proper logging/errors
        self.project_uuid = str(validator.enforce_uuid(self.project_uuid))
        self.job_uuid = str(validator.enforce_uuid(self.job_uuid))
        self.job_type = validator.enforce_enum(self.job_type, CstockJobType)
        self.job_inputs = validator.enforce_list_of_instances(
            self.job_inputs, CstockJobInput
        )



import logging

# TODO protocol instead of dependency
from .value_picker import JobInputValuePicker

from ceonmedia.core.project_input import CstockProjectInput
from ceonmedia.core.job_input import CstockJobInput
from ceonmedia.core.job_input import CstockJobInputType


logger = logging.getLogger(__name__)


# def create_job_input(
#     project_input: CstockProjectInput, value_picker: JobInputValuePicker
# ):
#     """
#     Create a job_input instance for the provided project_input
#     """
#     values = value_picker.get_values(project_input, num_values=5)
#     logger.debug(f"{value_picker} Got {len(values)} values:")
#     for value in values:
#         logger.debug(f"\t{value}")

#     return CstockJobInput(
#         name=project_input.name,
#         job_input_type=project_input.job_input_type,
#         values=values,
#     )



import logging
import json
from typing import Union
from typing import Optional
from pathlib import Path
from enum import Enum

from ceonmedia import CstockProject
from ceonmedia import CstockProjectInput
from ceonmedia import CstockJob
from ceonmedia import CstockJobType
from ceonmedia import CstockJobInput
from ceonmedia import CstockJobInputType
from ceonmedia.core.web_input import CstockWebInputType

from . import value_picker as vp

logger = logging.getLogger(__name__)


class FilePickerStrategy(Enum):
    ORDERED = "ordered"  # Get the first X found files
    RANDOM = "random"  # Randomly pick from all vailable files


# class FileSourceStrategy(Enum):
#     FOLDER_SETS = "folder_sets"  # List files inside a chosen folder
#     FILES = "files"  # List files that are not in a folder
#     ALL_FILES = "all_files"  # Create a list of all files, including subfolders


class DataPickerStrategy(Enum):
    FROM_FILE = "from_file"  # Read values from a json file on disk
    GENERATED = "generated"  # Random generation of values


class CstockJobGenerator:
    def __init__(
        self,
        # cstock_project: CstockProject,
        file_sources: Union[list[str], list[Path]],
        file_picker_strategy: Union[
            FilePickerStrategy, str
        ] = FilePickerStrategy.RANDOM,
        data_picker_strategy: Union[
            DataPickerStrategy, str
        ] = DataPickerStrategy.FROM_FILE,
        excluded_values: Optional[list] = None,
    ):
        """
        file_sources: A list of dirs containing files to be used as the job inputs.
        file_picker_strategy: The value_picker strategy used when choosing from files.
        data_picker_strategy: The value_picker strategy used when choosing from non-files.
        excluded_values: Values to be excluded from selection.
        """
        if not excluded_values:
            excluded_values = []
        # self.cstock_project = cstock_project
        if not isinstance(file_picker_strategy, FilePickerStrategy):
            file_picker_strategy = FilePickerStrategy(
                file_picker_strategy.lower()
            )
        if not isinstance(data_picker_strategy, DataPickerStrategy):
            data_picker_strategy = DataPickerStrategy(
                data_picker_strategy.lower()
            )

        self.file_sources = file_sources
        self.file_picker_strategy = file_picker_strategy
        self.data_picker_strategy = data_picker_strategy
        self.excluded_values = excluded_values

        # Maintain a list of value_pickers so that they can be used when/as intended
        self._value_pickers: dict[str, vp.CstockValuePicker] = {}

    def generate_job(
        self,
        cstock_project: CstockProject,
        job_type: CstockJobType = CstockJobType.PREVIEW,
    ) -> CstockJob:
        logger.warning(
            "TODO: Implement other file_picker_strategy (currently assumes to Random)"
        )
        job_inputs = []
        logger.debug(f"{self.file_sources=}")
        for project_input in cstock_project.project_inputs:
            logger.debug(f"Creating job_input for {project_input.name}")
            job_input = self.generate_job_input(project_input)
            logger.debug(f"Created job_input for {project_input.name}")
            job_inputs.append(job_input)

        logger.debug(f"Created {len(job_inputs)} job inputs:")
        job = CstockJob(
            job_inputs=job_inputs,
            project_uuid=str(cstock_project.project_uuid),
            job_type=job_type,
        )
        logger.debug(f"Created job:\n{job}")
        return job

    def generate_job_input(
        self, cstock_project_input: CstockProjectInput, num_entries=1
    ) -> CstockJobInput:
        logger.debug("Generating job input...")
        logger.warning(
            "TODO: Handle on_values_exhausted behaviour (error, cycle, repeat last value). (And make this a real generator with yeild?)"
        )
        value_picker = self._get_value_picker_for_project_input(
            cstock_project_input
        )
        chosen_values = value_picker.get_values(num_values=num_entries)
        logger.debug("Generated job input.")
        job_input = CstockJobInput(
            name=cstock_project_input.name,
            job_input_type=cstock_project_input.job_input_type,
            values=chosen_values,
        )
        return job_input

    def _get_value_picker_for_project_input(
        self, project_input: CstockProjectInput
    ) -> vp.CstockValuePicker:
        """Provides (and caches) a unique value picker for each project input"""
        value_picker = self._value_pickers.get(project_input.name)
        if value_picker:
            return value_picker

        logger.info(f"Creating new value picker for {project_input.name}")
        available_values = self._get_value_options_for_project_input(
            project_input
        )
        LOOKUP_STRATEGY = {
            FilePickerStrategy.ORDERED: vp.ValuePickerStrategy.ORDERED,
            FilePickerStrategy.RANDOM: vp.ValuePickerStrategy.RANDOM,
        }
        value_picker_strategy = LOOKUP_STRATEGY[self.file_picker_strategy]
        available_values_filtered = [
            value
            for value in available_values
            if value not in self.excluded_values
        ]
        new_value_picker = vp.CstockValuePicker(
            available_values_filtered,
            value_picker_strategy=value_picker_strategy,
        )
        logger.info(
            f"Created value picker for '{project_input.name}': {new_value_picker}"
        )
        self._value_pickers[project_input.name] = new_value_picker
        return new_value_picker

    def _get_value_options_for_project_input(
        self, project_input: CstockProjectInput
    ) -> list:
        """Handle the logic specific to this job_generator"""
        # Dropdowns provide their own values and will ignore the filesystem.
        logger.debug("Getting available values for job input generation...")
        if (
            project_input.web_input_type
            == CstockProjectInput.web_input_types.DROPDOWN
        ):
            available_options = _list_dropdown_values(project_input)
            return available_options

        if (
            project_input.job_input_type
            == CstockProjectInput.job_input_types.BOOL
        ):
            available_options = [False, True]
            return available_options

        # TODO implement different seleciton strategies.
        available_options = _get_value_options_from_files(
            self.file_sources, project_input
        )
        return available_options


def _list_dropdown_values(project_input: CstockProjectInput) -> list:
    """
    Extract the dropdown options from the project input and return the values as a list
    """
    if not project_input.web_input_type == CstockWebInputType.DROPDOWN:
        raise Exception(
            f"Cannot get dropdown options: Received project_input {project_input.name} is not a DROPDOWN web_input_type."
        )
    dropdown_options = project_input.type_settings["dropdown_options"]
    values = [option["value"] for option in dropdown_options]
    logger.debug(f"Got valid dropdown option values: {values}")
    return values


# def create_job_input(
#     project_input: CstockProjectInput, value_picker: vp.JobInputValuePicker
# ):
#     """
#     Create a job_input instance for the provided project_input
#     """
#     values = value_picker.get_values(project_input, num_values=5)
#     logger.debug(f"{value_picker} Got {len(values)} values:")
#     for value in values:
#         logger.debug(f"\t{value}")

#     return CstockJobInput(
#         name=project_input.name,
#         job_input_type=project_input.job_input_type,
#         values=values,
#     )


def _get_value_options_from_files(
    file_sources: Union[list[Path], list[str]],
    project_input: CstockProjectInput,
):
    """
    Return a list of available values for the current project_input.
    Values are sourced from self.file_sources, which are directories
    expected to follow an organizational structure.
        - File types are placed into folders.
        - Data types are stored in a json file with keys {arbitrary_name}: {value}
        (file_source_dir):
            doc/
                file1.docx.
                file2.docx.
            img/
                img1.jpg
                img2.png
            color.json <- {"white": "#AAAAAA", "black": "000000"}
    If multiple file_sources are provided, the available options will be combined.
    """
    if _is_file(project_input):
        # General case assumes the folder is named after the job_input_type
        fldr_name = project_input.job_input_type.lower()

        # Special cases, e.g. using only images generated from word docs
        # to match expected inputs for specific web_input_type
        if project_input.web_input_type == project_input.web_input_types.DOC:
            fldr_name = project_input.web_input_type.DOC.lower()

        files = []
        for file_source in file_sources:
            file_dir = Path(file_source, fldr_name)
            files += [file.name for file in _list_files_in_dir(file_dir)]
        logger.debug(f"Found {len(files)} from {len(file_sources)} sources.")
        return files

    # Read values from a json file
    possible_values = []
    json_filename = f"{project_input.job_input_type.lower()}.json"
    for file_source in file_sources:
        json_filepath = Path(file_source, json_filename)
        possible_values += _read_values_from_json_file(json_filepath)
    return possible_values


def _is_file(project_input) -> bool:
    file_types = [CstockJobInputType.IMG, CstockJobInputType.AUDIO]
    if project_input.job_input_type in file_types:
        return True
    return False


def _read_values_from_json_file(json_filepath: Union[Path, str]) -> list:
    """Read a json file on disk, return a list of values."""
    json_file = Path(json_filepath)
    if not json_file.is_file():
        raise FileNotFoundError(str(json_file))

    logger.debug(f"Reading json_file: {json_file}")
    with open(json_file) as f:
        json_str = f.read()
        data = json.loads(json_str)

    logger.debug("Loaded json data: ")
    for key, value in data.items():
        logger.debug(f"\t{key}: {value}")
    values = [value for value in data.values()]
    logger.debug(f"Got {len(values)} values from json file: ")
    for value in values:
        logger.debug(f"\t{value}")
    return values


def _list_files_in_dir(dir: Union[Path, str]) -> list[Path]:
    file_dir = Path(dir)
    if not file_dir.exists():
        raise NotADirectoryError(file_dir)
    filelist = file_dir.glob("**/*")
    files = [file for file in filelist if file.is_file()]
    logger.debug(f"Got {len(files)} files:")
    for file in files:
        logger.debug(f"\t{file}")
    return files



# Provides functionality for interacting with houdini parms.
# This module is ceonmedia AGNOSTIC.
import hou
from typing import List
from typing import Optional, Tuple


def get_visible_parms(parms):
    """
    Assumes all provides parms belong to the same node
    Checks if the parm is currently visible in the UI
    Returns only parms that are visible
    """
    parms[0].node().updateParmStates()
    visible_parms = [parm for parm in parms if not is_parm_hidden(parm)]
    return visible_parms


def create_parm(parm_data: dict, debug=False):
    # Takes a dictionary defining the parm to be created.
    # Example data:
    # {
    # "type": hou.FolderParmTemplate,
    # "name": "my_folder",
    # "label": "New Folder",
    # "folder_type": hou.folderType.Tabs,
    # "_children": [] # List of objects with same format as above
    # }
    if debug:
        print(f"{parm_data}\ncreate_parm got parm_data: ")
    parm_data = parm_data.copy()
    parm_type = parm_data.pop("type")
    if parm_type == hou.FolderParmTemplate:
        try:
            parm_data_list = parm_data.pop("_children")
        except KeyError:
            parm_data_list = []
        folder_parm = parm_type(**parm_data)
        for parm_data in parm_data_list:
            if debug:
                print("Creating child_parm from data: ", parm_data)
            child_parm = create_parm(parm_data)
            folder_parm.addParmTemplate(child_parm)
        return folder_parm
    else:
        return parm_type(**parm_data)


def print_templates_in_folder(node, folder_labels: tuple):
    ptg = node.parmTemplateGroup()
    # group_or_folder = ptg.findFolder(("Import", "Imported Data"))
    group_or_folder = ptg.findFolder(folder_labels)
    result = parm_templates_in_folder(group_or_folder)
    print("Found parms in folder {}: ".format(group_or_folder.name()))
    for parm in result:
        print("\t{}".format(parm.name()))
    return result


# Try not to use this. Only if necessary
def recook_scene():
    for node in hou.node("/").allSubChildren():
        node.cook(force=True)


def parm_templates_in_folder(
    group_or_folder, include_parms_in_inner_folders=False
):
    """Takes a parmTemplateGroup/folder as input and returns all input parms (including
    nested inside of other folders), excluding instances inside multiparm blocks.
    """
    to_return = []
    for parm_template in group_or_folder.parmTemplates():
        # Note that we don't want to return parm templates inside multiparm
        # blocks, so we verify that the folder parm template is actually
        # for a folder.
        to_return.append(parm_template)
        name = parm_template.name()
        print("got parm_template: {}".format(name))
        if (
            include_parms_in_inner_folders
            and parm_template.type() == hou.parmTemplateType.Folder
            and parm_template.isActualFolder()
        ):
            print(f"{name} type is folder AND isActualFolder!")
            for sub_parm_template in parm_templates_in_folder(parm_template):
                to_return.append(sub_parm_template)
                print(
                    f"Got sub_parm_template from {name}: {sub_parm_template.name()}"
                )
    return to_return


def parm_tuples_in_folder(
    node: hou.Node,
    folder_parm_template: hou.FolderParmTemplate,
    include_nested=False,
):
    # parm_templates_in_folder = folder_parm_template.parmTemplates()
    parm_templates = parm_templates_in_folder(
        folder_parm_template, include_parms_in_inner_folders=include_nested
    )
    parm_tuples_in_folder = [
        node.parmTuple(parm_template.name())
        for parm_template in parm_templates
    ]
    return parm_tuples_in_folder


def get_multiparm_parms(multiparm: hou.Parm, include_nested=False):
    parm_list = multiparm.multiParmInstances()
    if include_nested:
        for parm in parm_list:
            if is_multiparm(parm):
                parm_list += get_multiparm_parms(parm, include_nested=True)
    return parm_list


def get_multiparm_parms_grouped_by_instance(
    multiparm: hou.Parm, flatten_nested_multiparms: bool = False
):
    """If flatten_nested_multiparms is true, the children of any inner multiparms
    will be included. If false, the children are not included"""
    instance_parms = get_multiparm_parms(
        multiparm, include_nested=flatten_nested_multiparms
    )
    nest_depth = len(multiparm.multiParmInstanceIndices())
    # Group parms by instance
    instance_parms_grouped_by_index = {}
    for instance_parm in instance_parms:
        instance_num = instance_parm.multiParmInstanceIndices()[nest_depth]
        if str(instance_num) not in instance_parms_grouped_by_index.keys():
            instance_parms_grouped_by_index[str(instance_num)] = []
        instance_parms_grouped_by_index[str(instance_num)].append(
            instance_parm
        )

    return instance_parms_grouped_by_index


def create_color_parm(
    name: str,
    label: str,
    default_value: List[float] = [0.5, 0.25, 0.0],
    as_data: bool = False,
    **kwargs,
):
    parm_data = {
        "type": hou.FloatParmTemplate,
        "name": name,
        "label": label,
        "num_components": 3,
        "default_value": default_value,
        "look": hou.parmLook.ColorSquare,
        "naming_scheme": hou.parmNamingScheme.RGBA,
    }
    if kwargs:
        parm_data = {**parm_data, **kwargs}
    if as_data:
        return parm_data
    new_parm = create_parm(parm_data)
    return new_parm


def create_toggle_parm(
    name: str,
    label: str,
    default_value: bool = False,
    as_data: bool = False,
    **kwargs,
):
    parm_data = {
        "type": hou.ToggleParmTemplate,
        "name": name,
        "label": label,
        "default_value": default_value,
    }
    if kwargs:
        parm_data = {**parm_data, **kwargs}

    if as_data:
        return parm_data
    new_parm = create_parm(parm_data)
    return new_parm


def create_file_parm(
    name: str, label: str, default_value: str = "", as_data=False, **kwargs
):
    if default_value is None:
        default_value = f"$HIP/{label}.jpg"
    parm_data = {
        "type": hou.StringParmTemplate,
        "name": name,
        "label": label,
        "default_value": [default_value],
        "num_components": 1,
        "string_type": hou.stringParmType.FileReference,
        # "file_type": hou.fileType.Image,
    }
    if kwargs:
        parm_data = {**parm_data, **kwargs}
    if as_data:
        return parm_data
    return create_parm(parm_data)


def create_text_parm(
    name: str,
    label: str,
    default_value: str = "",
    multiline: bool = False,
    as_data: bool = False,
    **kwargs,
):
    parm_data = {
        "type": hou.StringParmTemplate,
        "name": name,
        "label": label,
        "default_value": [default_value],
        "num_components": 1,
        "string_type": hou.stringParmType.Regular,
        # "join_with_next": join_with_next,
    }
    if kwargs:
        parm_data = {**parm_data, **kwargs}
    if multiline:
        multiline_tag = {"editor": "1", "editorlines": "5-40"}
        current_tags = parm_data.get("tags", {})
        parm_data["tags"] = {**current_tags, **multiline_tag}
    if as_data:
        return parm_data
    return create_parm(parm_data)


def create_dropdown_parm(
    name: str,
    label: str,
    dropdown_options: dict,
    default_value: Optional[str] = None,
    as_data: bool = False,
    **kwargs,
):
    # dropdown_options = sorted(dropdown_options)
    menu_values = [
        str(dropdown_option["value"]) for dropdown_option in dropdown_options
    ]
    menu_labels = [
        str(dropdown_option["label"]) for dropdown_option in dropdown_options
    ]
    if default_value is None or default_value not in menu_values:
        default_value = menu_values[0]
    parm_data = {
        "type": hou.StringParmTemplate,
        "name": name,
        "label": label,
        "num_components": 1,
        "default_value": [default_value],
        "menu_items": (menu_values),
        "menu_labels": (menu_labels),
    }
    if kwargs:
        parm_data = {**parm_data, **kwargs}
    if as_data:
        return parm_data
    return create_parm(parm_data)


def create_int_parm(
    name: str,
    label: str,
    default_value: int = 0,
    as_data: bool = False,
    **kwargs,
):
    parm_data = {
        "type": hou.IntParmTemplate,
        "name": name,
        "label": label,
        "num_components": 1,
        "default_value": [int(default_value)],
    }
    if kwargs:
        parm_data = {**parm_data, **kwargs}
    if as_data:
        return parm_data
    return create_parm(parm_data)


def create_heading_parm(
    name: str, label: str, text: str, as_data=False, **kwargs
):
    if not text:
        text = ""
    parm_data = {
        "type": hou.LabelParmTemplate,
        "name": name,
        "label": label,
        "column_labels": [text],
    }
    if kwargs:
        parm_data = {**parm_data, **kwargs}
    heading_tag = {"sidefx::look": "heading"}
    current_tags = parm_data.get("tags", {})
    parm_data["tags"] = {**current_tags, **heading_tag}
    if as_data:
        return parm_data
    return create_parm(parm_data)


def create_message_parm(
    name: str, label: str, text: str, as_data=False, **kwargs
):
    if not text:
        text = ""
    parm_data = {
        "type": hou.LabelParmTemplate,
        "name": name,
        "label": label,
        "column_labels": [text],
    }
    if kwargs:
        parm_data = {**parm_data, **kwargs}
    block_tag = {"sidefx::look": "block"}
    current_tags = parm_data.get("tags", {})
    parm_data["tags"] = {**current_tags, **block_tag}
    if as_data:
        return parm_data
    return create_parm(parm_data)


def create_label_parm(
    name: str, label: str, text=None, as_data=False, **kwargs
):
    if not text:
        text = ""
    parm_data = {
        "type": hou.LabelParmTemplate,
        "name": name,
        "label": label,
        "column_labels": [text],
    }
    if kwargs:
        parm_data = {**parm_data, **kwargs}
    if as_data:
        return parm_data
    return create_parm(parm_data)


def create_folder_parm(
    name: str,
    label: str,
    child_parm_data: Optional[list] = None,
    folder_config: Optional[str] = None,
    as_data=False,
    **kwargs,
):
    folder_config_settings = {
        "collapsible": {"folder_type": hou.folderType.Collapsible},
        "borderless": {"tags": {"sidefx::look": "blank"}},
    }
    parm_data = {
        "type": hou.FolderParmTemplate,
        "name": name,
        "label": label,
        "default_value": 1,
    }
    if folder_config:
        # Get the folder config
        try:
            folder_settings = folder_config_settings[folder_config]
        except KeyError:
            raise Exception(
                "Invalid config string for hparms.create_folder_parm: `{folder_config}` not found"
            )
        parm_data = {**parm_data, **folder_settings}
    if kwargs:
        parm_data = {**parm_data, **kwargs}
    if child_parm_data:
        parm_data["_children"] = [parm_data for parm_data in child_parm_data]
    if as_data:
        return parm_data
    return create_parm(parm_data)


def create_multiparm_template(
    name: str,
    label: str,
    child_parm_data: Optional[list] = None,
    as_data=False,
    **kwargs,
):
    # Create the multiparm folder
    parm_data = {
        "type": hou.FolderParmTemplate,
        "name": name,
        "label": label,
        "default_value": 1,
        "folder_type": hou.folderType.MultiparmBlock,
    }
    if kwargs:
        parm_data = {**parm_data, **kwargs}
    if child_parm_data:
        parm_data["_children"] = [parm_data for parm_data in child_parm_data]
    if as_data:
        return parm_data
    return create_parm(parm_data)


def is_multiparm(parm: hou.Parm):
    if type(parm) == hou.Parm:  # is a 'real' parm, not a template
        parm_template = parm.parmTemplate()  # Get the parm template
    else:  # Is already a template
        parm_template = parm
    if (
        parm_template.type() == hou.parmTemplateType.Folder
        and not parm_template.isActualFolder()
    ):
        return True
    return False


def is_parm_hidden(parm: hou.Parm):
    # TODO finish this (doesn't work right now...)
    """Extend the default parm.isVisible() functionality to include a check if the parm
    is hidden due to a parent folder being hidden

    Args:
        parm (hou.Parm): The parameter to check the visibility of.

    Returns:
        bool: Whether the parameter is hidden in the Parameter
        interface.
    """
    is_hidden = False

    if not parm.isVisible():
        is_hidden = True
    # If the parm itself is not marked invisible, check parent folders to see if any of
    # those are hidden

    # TODO currently this doesn't work for parms that are inside of folders inside of
    # PLan:
    # from Node, getContainingFolders
    # from Node, get ptg
    # from ptg, find containing folder
    # use ptg is_folder_hidden to check for parent folder's visibility
    # ---
    # ptg = parm.node().parmTemplateGroup()
    # containing_folders = parm.containingFolders()
    # parent_multiparm = parm.parentMultiParm()
    # print("")
    # print(f"parm({parm.name()}) got:")
    # print(f"\tis_hidden: {is_hidden}")
    # print(f"\tcontaining_folders: {containing_folders}")
    # if containing_folders:
    # folder_parm = ptg.findFolder(containing_folders)
    # print(f"\tcontaining_folders parm: {folder_parm.name()}")
    # print(f"\tparent_multiparm: {parent_multiparm}")
    # print("ptg?: ", ptg)

    return is_hidden


def set_folder_parms(
    node: hou.Node, folder_label: Tuple[str, str], new_parms: List[hou.Parm]
):
    ptg = node.parmTemplateGroup()
    existing_fldr = ptg.findFolder(folder_label)
    # print("existing_fldr: ", existing_fldr)
    if not existing_fldr:
        # TODO missing folder exception
        raise Exception(
            f"set_folder_parms could not find folder label '{folder_label}' on node '{node.name()}'"
        )
    parms_tuple = tuple(new_parms)
    new_fldr_parm = hou.FolderParmTemplate(
        existing_fldr.name(),
        existing_fldr.label(),
        parms_tuple,
        folder_type=existing_fldr.folderType(),
        tags=existing_fldr.tags(),
    )
    # existing_fldr.setParmTemplates(parms_tuple)
    # Update parms in template group
    # for new_parm in new_parms:
    # fldr_parm.addParmTemplate(new_parm)

    # Push changes to node
    # Note: Does not change values if the parm already exists
    # (but will change the 'default value' of that parm)
    ptg.replace(existing_fldr.name(), new_fldr_parm)
    # existing_fldr.setParmTemplates(new_parms)
    node.setParmTemplateGroup(ptg)


def disable_all_parms_in_folder(node, folder_name: str):
    # TODO
    print("TODO: Disable parms in a folder")
    node.getFolder()


def disable_parms(node, parms: List, disabled=True):
    # print(f"Disabling parms on: {node.path()}")
    for parm in parms:
        print(f"Disabling parm: ({type(parm)=})")

        if isinstance(parm, hou.FolderParmTemplate):
            # If this is a FolderParmTemplate, we need to 'unpack it' to get the individual parms.
            # Then call this function recursively.
            print(f"Found folder parm: {parm.name()}")
            individual_parms = parm.parmTemplates()
            print(f"Containing {len(individual_parms)} parms")
            disable_parms(node, individual_parms)
            continue

        # node_parm = node.parm(parm.name())
        node_parm = node.parmTuple(parm.name())
        if node_parm:
            # print("locking parm: ", node_parm)
            # For some reason couldn't get parm.disable() to reflect in the UI
            # But lock works well for this case.
            node_parm.lock(disabled)
            # node_parm.disable(True)
        else:
            print(f"WARNING: Unable to disable, parm not foun: {parm.name()}")
            # raise Exception(
            #     "Unable to find expected parm: {}".format(parm.name())
            # )


def reset_parms_to_default(node: hou.Node, folder_name_tuple: Tuple[str, ...]):
    if not isinstance(folder_name_tuple, tuple):
        folder_name_tuple = (folder_name_tuple,)
    print(f"Setting values in {folder_name_tuple}: ")
    # node.parmTuple("cstock_user_color2_0").set((0.5, 0.5, 0.5))
    # Note: parmsInFolder (or parmTuplesInFolder) uses Label and NOT internal name
    # See docs for example usage
    # TODO check if folders exist on HDA, create if missing?
    try:
        existing_parms = node.parmsInFolder(folder_name_tuple)
    except hou.OperationFailed as e:
        print(f"parmsInFolder '{folder_name_tuple}' failed: {e}")
        # print("No existing parms found")
        existing_parms = []
    for parm in existing_parms:
        node.parm(parm.name()).revertToDefaults()
        value = node.parm(parm.name()).eval()
        print(
            "    Reverted {} to default value: {}".format(parm.name(), value)
        )


def all_nodes():
    nodes = hou.node("/").allSubChildren()
    return nodes


def get_dependant_nodes(node):
    """Returns all nodes that reference the target node"""
    print(f"Got dependents for node {node}: ")
    dependents = node.dependents()
    for dependant_node in dependents:
        print("\t", dependant_node)
    if not dependents:
        print("\t(None Found)")
    return dependents


def update_referencing_parms(node: hou.Node):
    # TODO make updates recursive? (in case parms are referencing the updated parm)
    print("")
    print(f"Evaluating {node.path()} referencing parms with parm.eval():")
    referencing_parms = node.parmsReferencingThis()
    print(f"Got {len(referencing_parms)} referencing parms.")

    # all_parms = node.parms()
    print(f"Got {len(referencing_parms)} referencing parms.")
    # print("referecing_parms: ", referencing_parms)
    for parm in referencing_parms:
        # print(f"Got parm: {parm}")
        evaluated_to = parm.eval()
        print("\t", parm, " : ", evaluated_to)
        references = parm.getReferencedParm()
        # evaluated_to = hou.evalParm(parm.path())
        print(f"\t\treferences: {references}")

    print("Cooking nodes: ")
    nodelist = []
    for parm in referencing_parms:
        parmnode = parm.node()
        if parmnode not in nodelist:
            parm.node().cook(force=True)
            print("Cooked: ", parmnode)
            nodelist.append(parmnode)



import logging

logger = logging.getLogger(__name__)


def log_dict_equality(orig_dict: dict, new_dict):
    logger.info(f"Checking dict equality...")
    for key in orig_dict:
        orig_value = orig_dict[key]
        new_value = new_dict[key]
        is_equal = orig_value == new_value
        logger.debug(f"Equality matched for key '{key}': {is_equal}")
        log_fn = logger.debug if is_equal else logger.warning
        log_fn(
            f"\nKey '{key}' old '==' new: {is_equal}:\
                \nOrig: ({type(orig_value)}){orig_value}\
                \nNew: ({type(new_value)}){new_value}\
            "
        )



import logging
import inspect
from typing import Optional, Type, List
from ceonmedia.log import handlers

# For logging of the logging creation process
self_logger = logging.getLogger(__name__)
self_logger.setLevel("INFO")
self_logger.info(f"Self_logger {self_logger.name} initialized!")


def ceonmedia_logger():
    """The parent logger for the ceonmedia package"""
    return logging.getLogger("ceonmedia")


def create_new_logger(
    logger_name: str,
    level: Optional[str] = None,
    handlers: Optional[List[logging.Handler]] = None,
):
    """
    Create a new logger.
    Raise an exception if the logger already exists.
    """
    if not handlers:
        handlers = []
    if logger_name in logging.Logger.manager.loggerDict.keys():
        raise Exception(f"Cannot create logger. Logger already exists: {logger_name}")
    # default_output_log = "logs/logs.log"
    self_logger.debug(f"Creating logger: {logger_name}")
    new_logger = logging.getLogger(logger_name)
    if level:
        new_logger.setLevel(level)
    for handler in handlers:
        new_logger.addHandler(handler)
    return new_logger


def module_logger(
    *, level: Optional[str] = None, handlers: Optional[List[logging.Handler]] = None
) -> logging.Logger:
    """
    Creates a logger with a name that matches the module name of the caller
    For example, when inside of
        ceonmedia.core.job_input
    instead of creating a logger and using __name__, we can instead just use
    module_logger() and __name__ will be inherited automatically based
    on where module_logger() was called from.
    """
    self_logger.debug("TEST creating module logger")

    # Get the caller's module name. This saves having to pass __name__ as an arg
    # each time when setting up the logger
    caller_frame = inspect.stack()[1]
    caller_module = inspect.getmodule(caller_frame[0])
    if not caller_module:
        return logging.getLogger("_unknown_module_logger_")
    caller_module_name = caller_module.__name__
    self_logger.debug(f"Caller: {caller_module_name}")

    module_logger = create_new_logger(
        caller_module_name, level=level, handlers=handlers
    )
    self_logger.info(f"Creating module logger: {module_logger.name}")
    return module_logger



from ceonmedia import CstockJobInputType, CstockWebInputType

# Examples that should always work
SUCCESS_CASES = [
    {
        "name": "myTestProjInput",
        "job_input_type": CstockJobInputType.IMG,
        "web_input_type": CstockWebInputType.IMG,
    }
]

# Examples that we know should not work.
FAIL_CASES = [
    {  # Fail unrecognized kwargs to alert user of possible unexpected data inputs if unpacking a dict
        "name": "myTestProjInput",
        "job_input_type": CstockJobInputType.IMG,
        "web_input_type": CstockWebInputType.IMG,
        "unknownKawrgNotRecognized": [1, 2, 3],
    },
    {  # Passed the wrong Enum type to job_input (the passed enum value 'DROPDOWN' doesn't exist on the expected enum)
        "name": "myTestProjInput",
        "job_input_type": CstockWebInputType.DROPDOWN,
        "web_input_type": CstockWebInputType.IMG,
    },
    {  # Passed the wrong Enum type to job_input (but the passed enum value does also exist on the expected enum)
        "name": "myTestProjInput",
        "job_input_type": CstockWebInputType.IMG,
        "web_input_type": CstockWebInputType.IMG,
    },
]



from uuid import uuid4
from ceonmedia.core import CstockRenderTask
from ceonmedia.core import render_task
from ceonmedia.core.file_reference import CstockFileReference
from ceonmedia.core.file_reference import CstockFileSourceType
from ceonmedia import CstockRenderPipeline

RENDER_TASKS = [
    CstockRenderTask(
        task_name="render_the_frames",
        task_output="output_file.$r.exr",
        task_input=CstockFileReference("hou_project/myhipfile.hip"),
        app_type=render_task.CstockRenderAppType.HOU,
        app_render_settings=render_task.CstockRenderTaskAppSettingsHou(
            frames="0 10", target_node="/out/karma1"
        ),
        app_version="20.0",
    ),
    CstockRenderTask(
        task_name="ffmpeg_frames_to_mov",
        task_output="my_output.mp4",
        task_input=CstockFileReference(
            "in_file.txt", file_source=CstockFileSourceType.JOB_INPUT
        ),
        app_version="5",
        app_type=render_task.CstockRenderAppType.FFMPEG,
        app_render_settings=render_task.CstockRenderTaskAppSettingsFFMPEG(
            input_args="",
            output_args="",
        ),
    ),
]

SUCCESS_CASES = [
    {  # Expected case
        "render_tasks": RENDER_TASKS,
        "output_task": "ffmpeg_frames_to_mov",
    },
]

FAIL_CASES = [
    {  # No render tasks
        "render_tasks": [],
        "output_task": "ffmpeg_frames_to_mov",
    },
    {  # Primary output task name is not an existing task
        "render_tasks": RENDER_TASKS,
        "output_task": "a_task_name_that_deosnt_exit",
    },
]



import logging
from typing import Type
from copy import deepcopy
from dataclasses import asdict

from ceonmedia.core.base import CstockBaseClass

logger = logging.getLogger(__name__)
# logger.setLevel("INFO")


def test_dict_to_cstock_object(dict_data, cstock_cls: Type[CstockBaseClass]):
    """Test that a dict is cstock loadable
    in both directions"""
    print()
    logger.debug("\ndict_data / to_cstock / back_to_dict ")
    print(f"\t{dict_data}")
    new_cstock_object = cstock_cls(**dict_data)
    print(f"\t{new_cstock_object}")
    back_to_dict = deepcopy(asdict(new_cstock_object))
    print(f"\t{back_to_dict}")
    assert dict_data == back_to_dict


def test_cstock_object_to_dict(cstock_object: CstockBaseClass):
    """
    Convert the cstock_object to a dict.
    Confirm that the created dict can successfully reinstantiate the object
    """
    dict_data = deepcopy(asdict(cstock_object))
    logger.debug("")
    logger.debug("original / dict / back_to_cstock ")
    logger.debug(f"\t{cstock_object}")
    logger.debug(f"\t{dict_data}")
    cstock_from_dict = cstock_object.__class__(**dict_data)
    logger.debug(f"\t{cstock_from_dict}")
    assert asdict(cstock_object) == asdict(cstock_from_dict)


def run_dict_serialization_tests(cstock_object) -> None:
    """Tests conversion to/from class instance and dict"""
    # Dict data
    logger.info("Testing dict/cstock workflows...")
    dict_data = deepcopy(asdict(cstock_object))
    test_cstock_object_to_dict(cstock_object)
    test_dict_to_cstock_object(dict_data, cstock_object.__class__)



#  https://realpython.com/python-json/
import json
import logging
from typing import Type, Callable, Dict, List
from functools import partial
from uuid import UUID

from ceonmedia.core.base import CstockBaseClass
from ceonmedia import (
    CstockProjInput,
    CstockProjectInfo,
    CstockRenderTask,
    CstockJobInput,
    CstockRenderPipeline,
)
from ceonmedia.log import printify
from ceonmedia import errors

logger = logging.getLogger(__name__)


def load_from_list(
    list_of_objects: List[dict], cstock_cls: Type[CstockBaseClass]
) -> List[CstockBaseClass]:
    return [cstock_cls(**object) for object in list_of_objects]


def load_from_dict(dict_from_json: dict, cstock_cls: Type[CstockBaseClass]):
    logger.info(f"Loading {cstock_cls.__name__} with dict: ")
    logger.info(f"{dict_from_json}")
    try:
        return cstock_cls(**dict_from_json)
    except TypeError as e:
        raise errors.CstockInstantiationError(
            cstock_cls=cstock_cls,
            errors=[str(e), f"Received:{printify(dict_from_json)}"],
        )


KEY_ACTION_MAP = {  # Special actions to be taken if a particular key is found
    "project_info": partial(load_from_dict, cstock_cls=CstockProjectInfo),
    "project_inputs": partial(load_from_list, cstock_cls=CstockProjInput),
    "job_inputs": partial(load_from_list, cstock_cls=CstockJobInput),
    "render_tasks": partial(load_from_list, cstock_cls=CstockRenderTask),
    "render_pipeline": partial(load_from_dict, cstock_cls=CstockRenderPipeline),
    "uuid": UUID,
    "task_uuid": UUID,
}


def decode_complex(dct):
    """
    Example of custom decoding logic.
    In this case a dict of
    {
        "__complex__": true,
        "real": 42,
        "imag": 36
    }
    Could be interpreted and loaded as a complex number object (python builtin) with:
    if "__complex__" in dct:
        return complex(dct["real"], dct["imag"])
    """
    # If a special key is detected, process the key's value with the appropriate function
    for key in dct.keys():
        if key in KEY_ACTION_MAP.keys():
            action_fn = KEY_ACTION_MAP[key]
            logger.debug(f"Found special key '{key}', running action: {action_fn}")
            dct[key] = action_fn(dct[key])
    return dct


LOAD_ARGS = {"object_hook": decode_complex}


def loads(json_string: str):
    return json.loads(json_string, **LOAD_ARGS)


def load(filepath):  # TODO
    return json.load(filepath, **LOAD_ARGS)



from .job import CstockJob
from .job import CstockJobType
from .job_input import CstockJobInput
from .job_input import CstockJobInputType
from .web_input import CstockWebInputType

from .project_input import CstockProjectInput
from .project import CstockProject

from .file_reference import CstockFileReference
from .file_reference import CstockFileSourceType

# from .render import CstockRenderPipeline

# from .render import RenderOutputSettings
from .render import Dimensions

# from .render_task import CstockRenderAppType
# from .render_task import CstockRenderTask

from .submission_input import CstockSubmissionInput



# The UI for setting up a single project input.
import customtkinter
import tkinter
from ceonmedia import CstockJobInputType
from ceonmedia import CstockWebInputType
from ceonmedia import CstockProjectInput
from typing import List
from typing import Optional


def dropdown_callback(arg):
    print(arg)


def valid_ui_types(
    job_input_type: CstockJobInputType,
) -> List[CstockWebInputType]:
    LOOKUP = {
        CstockJobInputType.AUDIO: [CstockWebInputType.AUDIO],
        CstockJobInputType.COLOR: [CstockWebInputType.COLOR],
        # CstockJobInputType.FLOAT: [CstockWebInputType.FLOAT], # TODO add float webinput?
        CstockJobInputType.IMG: [
            CstockWebInputType.IMG,
            CstockWebInputType.DOC,
        ],
        CstockJobInputType.INT: [CstockWebInputType.INT],
        CstockJobInputType.STRING: [CstockWebInputType.TEXT],
    }
    return LOOKUP[job_input_type]


LABEL_STICKY = "e"


def test_callback(value):
    print(value)


def format_project_name(value):
    """Format the name according to requirements"""
    out_value = value
    out_value = out_value.replace(" ", "_")
    out_value = out_value.lower()
    # print("out: ", out_value)
    return out_value


default_project_input = CstockProjectInput(
    name="Unnamed Project Input",
    job_input_type=CstockJobInputType.IMG,
    web_input_type=CstockWebInputType.IMG,
)


class FrameProjectInput(customtkinter.CTkFrame):
    def __init__(
        self, master, cstock_project_input: Optional[CstockProjectInput] = None
    ):
        super().__init__(master)
        # self.grid_columnconfigure((0, 1), weight=1)
        self.grid_columnconfigure(0)
        self.grid_columnconfigure(1, weight=2)

        if not cstock_project_input:
            cstock_project_input = default_project_input

        print("Creating frame with cstock_project_input: ")
        print(cstock_project_input)

        # Label shown in the UI
        self.title = customtkinter.CTkLabel(
            self,
            text=cstock_project_input.name,
            fg_color="gray30",
            font=("", 18),
            corner_radius=6,
        )
        self.title.grid(
            row=0, column=0, padx=10, pady=(10, 0), sticky="ew", columnspan=2
        )
        # Delete button
        self.button_delete = customtkinter.CTkButton(
            self,
            width=30,
            text="X",
            fg_color="#a44",
            font=("", 18),
            command=self.destroy,
        )
        self.button_delete.grid(
            row=0, column=0, padx=10, pady=(10, 0), sticky="w"
        )

        def on_name_change(*args):
            value = f"{self.entry_job_input_name.get()}"
            value = format_project_name(value)
            self.title.configure(text=value)
            # print(f"{value=}")

        # Project Input Name
        row = 1
        self.label_job_input_name = customtkinter.CTkLabel(
            self, text="Project Input Name"
        )
        self.label_job_input_name.grid(
            row=row, column=0, padx=(20, 0), sticky=LABEL_STICKY
        )
        # Use a variable to trigger onchange callback
        self.name_var = tkinter.StringVar()
        self.name_var.trace_add("write", on_name_change)
        self.entry_job_input_name = customtkinter.CTkEntry(
            self, placeholder_text="job_input_name", textvariable=self.name_var
        )
        self.entry_job_input_name.grid(
            row=row, column=1, padx=20, pady=5, sticky="ew"
        )

        # Project Input Description
        row += 1
        self.label_project_input_description = customtkinter.CTkLabel(
            self, text="Project Input Description"
        )
        self.label_project_input_description.grid(
            row=row, column=0, padx=(20, 0), sticky=LABEL_STICKY
        )
        self.textbox_project_input_description = customtkinter.CTkTextbox(
            self, height=76
        )
        self.textbox_project_input_description.grid(
            row=row, column=1, padx=20, pady=5, sticky="ew"
        )

        # Job Input Type
        row += 1
        self.label_job_input_type = customtkinter.CTkLabel(
            self, text="Job Input Type"
        )
        self.label_job_input_type.grid(
            row=row, column=0, padx=(20, 0), sticky=LABEL_STICKY
        )
        options_job_input_type = [e.value for e in CstockJobInputType]
        self.dropdown_job_input_type = customtkinter.CTkOptionMenu(
            self,
            values=options_job_input_type,
            command=self.refresh_web_ui_options,
        )
        self.dropdown_job_input_type.grid(
            row=row, column=1, padx=20, pady=5, sticky="ew"
        )

        # Web Input Type
        row += 1
        self.label_web_input_type = customtkinter.CTkLabel(
            self, text="Web Input Type"
        )
        self.label_web_input_type.grid(
            row=row, column=0, padx=(20, 0), sticky=LABEL_STICKY
        )
        options_web_ui = valid_ui_types(options_job_input_type[0])
        self.dropdown_web_input_type = customtkinter.CTkOptionMenu(
            self,
            values=options_web_ui,
            command=dropdown_callback,
        )
        self.dropdown_web_input_type.grid(
            row=row, column=1, padx=20, pady=5, sticky="ew"
        )

        # Set default values
        self.set(cstock_project_input)

    def refresh_web_ui_options(self, job_input_type_value: str):
        job_input_type = CstockJobInputType(job_input_type_value)
        web_ui_types = valid_ui_types(job_input_type)
        options = [web_ui_type.value for web_ui_type in web_ui_types]
        self.dropdown_web_input_type.configure(values=options)
        self.dropdown_web_input_type.set(options[0])
        if len(options) <= 1:
            self.dropdown_web_input_type.configure(state="disabled")
        else:
            self.dropdown_web_input_type.configure(state="normal")

    def set(self, cstock_project_input: CstockProjectInput):
        # Set values
        self.textbox_project_input_description.insert(
            "0.0", cstock_project_input.description
        )
        self.dropdown_job_input_type.set(
            cstock_project_input.job_input_type.value
        )
        self.name_var.set(cstock_project_input.name)
        self.refresh_web_ui_options(cstock_project_input.job_input_type.value)

    def get(self) -> CstockProjectInput:
        """Build a CstockprojectInput instance from the provided settings in the UI"""
        name = self.entry_job_input_name.get()
        job_input_type = self.dropdown_job_input_type.get()
        web_input_type = self.dropdown_web_input_type.get()
        description = self.textbox_project_input_description.get("0.0", "end")
        description = description.strip()  # Remove trailing newline
        cstock_project_input = CstockProjectInput(
            name,
            job_input_type=job_input_type,
            web_input_type=web_input_type,
            description=description,
        )
        print(name, job_input_type, web_input_type, description)
        print(cstock_project_input)
        return cstock_project_input



import logging
from pathlib import Path
from ceonmedia.log import formatters

self_logger = logging.getLogger(__name__)


def file_handler(
    log_filepath: Path, formatter: logging.Formatter = formatters.PROD, level=None
):
    self_logger.info(f"Got log_filepath: {log_filepath}")
    log_dir = Path(log_filepath.parent)
    self_logger.debug(f"log_dir: {log_dir}")
    file_handler = logging.FileHandler(log_filepath)
    file_handler.setFormatter(formatter)
    if level:
        file_handler.setLevel(level)
    return file_handler


def stream_handler(formatter: logging.Formatter = formatters.DEV, level=None):
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    if level:
        stream_handler.setLevel(level)
    return stream_handler


def create_dir(dir: Path):
    self_logger.debug(f"dir.exists(): {dir.exists()}")
    if not dir.exists():
        self_logger.info(f"Creating new folder(s): {dir}")
        dir.mkdir(parents=True, exist_ok=True)
    else:
        self_logger.info(f"dir already exists: {dir}")



import logging
from typing import Union
from pathlib import Path

import pytest

from ceonmedia import filesystem
from ceonmedia import errors

logger = logging.getLogger(__name__)

# Folders that store test projects
THIS_DIR = Path(__file__).parent
PROJECTS_DIR_SUCCESS = f"{THIS_DIR}/local_projects/success"
PROJECTS_DIR_FAIL = f"{THIS_DIR}/local_projects/fail"

# A single chosen project to test class functions
PROJECT_PATH = Path(PROJECTS_DIR_SUCCESS, "Waving Flag")


def list_folders_in_dir(directory: Union[str, Path]):
    directory = Path(directory)  # Cast string to Path
    return [entry for entry in directory.iterdir() if entry.is_dir()]


def test_local_project_load():
    local_project = filesystem.CstockLocalProject(PROJECT_PATH)
    cstock_project = local_project.load()
    logger.info(f"Loaded CstockProject from local_project:\n{cstock_project=}")
    # TODO validate project data?


def test_local_project_load_success():
    source_dir = PROJECTS_DIR_SUCCESS
    local_project_dirs = list_folders_in_dir(source_dir)
    if not local_project_dirs:
        raise Exception(f"Success cases not tested: No projects found in {source_dir}")
    for local_project_dir in local_project_dirs:
        logger.info(f"Testing local project success case: {local_project_dir}")
        local_project = filesystem.CstockLocalProject(local_project_dir)
        cstock_project = local_project.load()
        logger.info(f"Loaded CstockProject from local_project:\n{cstock_project=}")


def test_local_project_load_fail_cases():
    source_dir = PROJECTS_DIR_FAIL
    local_project_dirs = list_folders_in_dir(source_dir)
    if not local_project_dirs:
        raise Exception(f"Failure cases not tested: No projects found in {source_dir}")
    for local_project_dir in local_project_dirs:
        logger.info(f"Testing local project fail case: {local_project_dir}")
        with pytest.raises((errors.CstockInstantiationError, FileNotFoundError, NotADirectoryError)):
            local_project = filesystem.CstockLocalProject(local_project_dir)
            cstock_project = local_project.load()



import logging
import math
from typing import Optional
from datetime import timedelta
from dataclasses import dataclass
from typing import Union
from pathlib import Path

from ceonmedia import config
from ceonmedia import CstockJob
from ceonmedia import CstockJobInput

logger = logging.getLogger(__name__)


# TODO move vtt file to cstock filesystem module
@dataclass
class VTTFileCue:
    """A single 'entry' in a VTT file"""

    identifier: str
    start_time_seconds: float
    end_time_seconds: float
    content: str

    def apply_time_offset(self, time_seconds: float):
        self.start_time_seconds += time_seconds
        self.end_time_seconds += time_seconds

    def __str__(self):
        return (
            f"{self.identifier}\n"
            f"{seconds_to_timestamp(self.start_time_seconds)}"
            f" --> {seconds_to_timestamp(self.end_time_seconds)}\n"
            f"{self.content}"
        )


class VTTFileProjectInputExample:
    def __init__(
        self,
        file_path: Union[Path, str],
        project_input_name: str,
        file_description: str = "",
    ):
        """
        start_frame: The frame number that corresponds to time 0.00 in the video.
        """
        self.file_path = Path(file_path)
        self.project_input_name = project_input_name
        if not file_description:
            file_description = (
                f"ceonmedia project input examples: {self.project_input_name}"
            )
        self.file_description = file_description
        self._cues: list[VTTFileCue] = []

    def save(
        self,
        ceonmedia_jobs: list[CstockJob],
        project_video_length_seconds: Optional[float] = None,
        project_start_frame: Optional[int] = None,
    ) -> Path:
        """Generate the vtt file for the provided ceonmedia_jobs."""
        # cues = self._create_cues_for_cstock_jobs(ceonmedia_jobs)
        if not project_start_frame:
            logger.warning(
                "Did not receive project_start_frame argument. Getting minimum value from job_limits"
            )
            job_start_frames = [
                int(ceonmedia_job.limit_frames.split(" ")[0])
                for ceonmedia_job in ceonmedia_jobs
                if ceonmedia_job.limit_frames
            ]
            lowest_frame = min(job_start_frames)
            project_start_frame = lowest_frame
            logger.warning(f"{project_start_frame=}")
            # project_start_frame = #TODO get min frame.

        cues = [
            self._create_cue_for_cstock_job(
                job, str(i), start_frame=project_start_frame
            )
            for i, job in enumerate(ceonmedia_jobs)
        ]
        logger.debug(f"Created cues: {cues}")
        for cue in cues:
            logger.debug(cue)
        offset_overlapping_cues(cues, project_video_length_seconds)
        logger.debug(f"Updated? cues: {cues}")
        for cue in cues:
            logger.debug(cue)
        self._cues = cues
        file_content = self._file_content()
        if not self.file_path.parent.exists():
            logger.info(f"Creating dir: {self.file_path.parent}")
            self.file_path.parent.mkdir(exist_ok=True, parents=True)
        logger.info(f"Saving file: {self.file_path}")
        with open(self.file_path, "w") as f:  #
            f.write(file_content)
        return Path(self.file_path)

    def _create_cue_for_cstock_job(
        self,
        ceonmedia_job: CstockJob,
        cue_identifier: str,
        fps=config.DEFAULT_FPS,
        start_frame: int = 0,
    ) -> VTTFileCue:
        """
        start_frame: The frame which corresponds to time 0.0 in the video.
        """
        cue_identifier = str(cue_identifier)
        if not ceonmedia_job.limit_frames:
            raise Exception(
                "VTT file received a ceonmedia job which does not have limit_frames set."
            )
        # TODO refactor limit_frames (and cstock workflow as a whole) to accept
        # time ranges instead of frames?
        job_start_frame = int(ceonmedia_job.limit_frames.split(" ")[0])
        job_end_frame = int(ceonmedia_job.limit_frames.split(" ")[1])
        start_time = (job_start_frame - start_frame) / fps
        end_time = (job_end_frame - start_frame + 1) / fps
        # Add +1 to the end frame so that the end time represents the END of the frame.
        # For example, first frames starts at time 0.0 but will show onscreen for 1/fps seconds.
        # To make the time ranges fully encompass the frame range, we need to capture the END of the end
        # frame. This also causes the cue timings to be joined end-to-start of next cue.
        target_job_input: Optional[CstockJobInput] = None
        for job_input in ceonmedia_job.job_inputs:
            if job_input.name == self.project_input_name:
                target_job_input = job_input
                break
        if not target_job_input:
            raise Exception(
                f"Could not find job_input with project_input_name '{self.project_input_name}'"
            )
        values = [str(value) for value in target_job_input.values]
        content = "\n".join(values)
        cue = VTTFileCue(
            identifier=str(cue_identifier),
            start_time_seconds=start_time,
            end_time_seconds=end_time,
            content=content,
        )
        return cue

    def cues(self):
        return self._cues

    def _header(self) -> str:
        vtt_header = "WEBVTT"  # Required start of file string for VTT files.
        if self.file_description:
            if "-->" not in self.file_description:
                vtt_header += f" {self.file_description}"
            else:
                logger.warning(
                    "File description contains invalid content '-->', it will not be added to the vtt file."
                )
        return vtt_header

    def _file_content(self) -> str:
        """Return the string representing the full VTT file"""
        logger.debug(f"Formatting VTT content")
        vtt_data_body = "\n\n".join([str(cue) for cue in self.cues()])
        vtt_data = self._header() + "\n\n" + vtt_data_body
        return vtt_data


def offset_overlapping_cues(
    cues: list[VTTFileCue],
    project_video_length_seconds: Optional[float] = None,
):
    """
    When we loop/repeat a cstock project video with varying example inputs, the generated job frameranges
    fall out of sync with the total video length.
    If we find a timestamp that overlaps with a previous timestimp, push it to the next 'video cycle'
    """
    # NOTE: This logic assumes that cues are indexed in order of appearance in the video
    if not cues:
        logger.warning(
            "Could not apply video cycling to vtt timestamps: No cues received."
        )
        return cues

    if not project_video_length_seconds:
        logger.warning(
            f"Did not receive explicit declaration of project length. Assuming latest found cue end time as"
            "the project length. This may cause desync if cues were not created for the full video length."
        )
        max_cue_time = max(
            cues, key=lambda x: x.end_time_seconds
        ).end_time_seconds
        logger.debug(f"{max_cue_time=}")
        project_video_length_seconds = max_cue_time
        logger.debug(
            f"Assuming video length from existing cues: {project_video_length_seconds=}"
        )

    def _relative_time(time_seconds: float) -> float:
        """Return the time relative to the current loop"""
        if time_seconds == 0:
            return 0  # 0 is always the START of a loop.

        relative_time = time_seconds % project_video_length_seconds
        if relative_time == 0:
            # Since we already handled the 0 case, we can assume we are at the END of the loop.
            return project_video_length_seconds
        return relative_time

    previous_cues_max_end_time = 0.0
    for cue in cues:
        logger.debug(f"{previous_cues_max_end_time=}")
        if cue.start_time_seconds >= previous_cues_max_end_time:
            # The cues are progressing in normal expected order
            previous_cues_max_end_time = max(
                previous_cues_max_end_time, cue.end_time_seconds
            )
            logger.debug(f"cue({cue.identifier}) does not need shifting.")
            continue

        # This cue is starting earlier than a previous cue
        # Determine if we need move to the next loop or append the cue to the current loop
        relative_start_time = _relative_time(cue.start_time_seconds)
        previous_cues_max_end_time_relative = _relative_time(
            previous_cues_max_end_time
        )
        logger.debug(f"{relative_start_time=}")
        logger.debug(f"{previous_cues_max_end_time_relative=}")
        current_loop = _calculate_loop_num_for_time(
            previous_cues_max_end_time, project_video_length_seconds
        )
        logger.debug(f"{current_loop=}")
        if relative_start_time >= previous_cues_max_end_time_relative:
            target_loop = current_loop
        else:
            target_loop = current_loop + 1
        logger.debug(f"{target_loop=}")

        set_cue_to_loop_num(cue, target_loop, project_video_length_seconds)
        previous_cues_max_end_time = max(
            previous_cues_max_end_time, cue.end_time_seconds
        )


def seconds_to_timestamp(time_seconds: float) -> str:
    """Return a timestamp formatted for a VTT file"""
    td = timedelta(seconds=time_seconds)
    # Get the hours, minutes, and seconds.
    minutes, seconds = divmod(td.seconds, 60)
    hours, minutes = divmod(minutes, 60)
    # Round the microseconds to millis.
    millis = int(round(td.microseconds / 1000, 0))
    str_timestamp = f"{hours:02}:{minutes:02}:{seconds:02}.{millis:03}"
    return str_timestamp


def _calculate_loop_num_for_time(
    time_seconds: float, loop_length_seconds: float
) -> int:
    """
    Return the loop number based on the provided time.
    0 represents the videos first playthrough
    1 represents the first repeat of the video
    """
    loop_num = math.floor(time_seconds / loop_length_seconds)
    if time_seconds % loop_length_seconds == 0:
        # Special case in which the cue ends exactly at the end time of the project
        return loop_num - 1
    return loop_num


def set_cue_to_loop_num(
    cue: VTTFileCue, loop_num: int, loop_length_seconds: float
):
    """
    Loop num 0 is the first video playthrough.
    Loop num 1 is the first repeat of the video.
    """
    # In case the timestamp was already outside of the project length, use the remainder
    # to calcualte the relative start time.
    time_seconds = cue.start_time_seconds
    relative_start_time = time_seconds % loop_length_seconds
    loop_num_start_time_seconds = loop_length_seconds * loop_num
    new_start_time = loop_num_start_time_seconds + relative_start_time
    offset_amount = new_start_time - time_seconds

    # Apply offset to cue
    logger.debug(
        f"Current cue time: {cue.start_time_seconds} -> {cue.end_time_seconds})"
    )
    cue.apply_time_offset(offset_amount)
    logger.debug(
        f"Updated cue time: {cue.start_time_seconds} -> {cue.end_time_seconds})"
    )
    logger.debug(
        f"Pushed cue({cue.identifier}) to video cyle {loop_num}: {cue.start_time_seconds} -> {cue.end_time_seconds}"
    )



from __future__ import annotations
import logging
from typing import List, Optional, ClassVar, Type
from dataclasses import dataclass, field

import ceonmedia.errors as errors
from ceonmedia.core.base import CstockBaseClass, CstockBaseEnum
from ceonmedia.core import validator
from ceonmedia.log.diagnostics import init_args

# Don't use StrEnum  until Houdini and all other apps are updated to Python 3.11
# from enum import StrEnum, auto
from enum import Enum

logger = logging.getLogger(__name__)


# Inherting from str makes the enum JSON serializable
class CstockJobInputType(str, Enum):
    AUDIO = "audio"
    BOOL = "bool"
    COLOR = "color"
    FLOAT = "float"
    IMG = "img"
    INT = "int"
    STRING = "string"


FILE_TYPES = [CstockJobInputType.AUDIO, CstockJobInputType.IMG]


@dataclass
class CstockJobInput(CstockBaseClass):
    """Represents a single Job Input."""

    #  Make types easily accessible for user without needing to explicitly import CstockJobInputType
    types: ClassVar[Type[CstockJobInputType]] = CstockJobInputType

    name: str
    job_input_type: CstockJobInputType
    values: List = field(default_factory=list)
    # Computed field but cannot use init=False becasue it causes an error when loading from json import.
    is_file_type: bool = False

    def __post_init__(self):
        logger.debug(
            f"Instantiating {self.__class__.__name__} with __init__ locals: {init_args()}"
        )
        self.job_input_type = validator.enforce_enum(
            self.job_input_type, CstockJobInputType
        )
        self.is_file_type = self.job_input_type in FILE_TYPES
        self.validate()
        logger.debug(f"Initialized {self}")

    def __str__(self):
        msg = f"<{self.__class__.__name__}({self.job_input_type}) '{self.name}': {self.values}>"
        return msg

    def __repr__(self):
        return self.__str__()

    def validate(self):
        """Performs validation, collecting errors so that they can all be reported together"""
        logger.debug(f"Validating {self.__class__.__name__} '{self.name}'...")
        validation_errors = []
        if not isinstance(self.values, list):
            validation_errors.append(
                f"{self.__class__.__name__}.values expects a list but received type: {type(self.values)}"
            )

        # Raise errors
        if validation_errors:
            raise errors.CstockInstantiationError(
                cstock_cls=self.__class__, errors=validation_errors
            )
        logger.debug(f"... Validated {self}")



from __future__ import annotations
import logging
from dataclasses import dataclass, field
from typing import ClassVar, Type

from ceonmedia.core.web_input import CstockWebInputType
from ceonmedia.core.job_input import CstockJobInputType
from ceonmedia.core.base import CstockBaseClass
from ceonmedia.core import validator
from ceonmedia.log import printify
from ceonmedia import errors as cstock_errors

logger = logging.getLogger(__name__)

HARD_LIMIT_NUM_ENTRIES_MAX = 10


@dataclass
class CstockProjectInput(CstockBaseClass):
    """Represents a single Project Input."""

    #  Make types easily accessible for user without needing to explicitly import CstockJobInputType
    job_input_types: ClassVar[Type[CstockJobInputType]] = CstockJobInputType
    web_input_types: ClassVar[Type[CstockWebInputType]] = CstockWebInputType

    name: str
    job_input_type: CstockJobInputType
    web_input_type: CstockWebInputType
    description: str = ""
    type_settings: dict = field(
        default_factory=dict
    )  # User-provided arguments that are unique for this type, e.g. "dropdown_options"
    num_entries_min: int = 1
    num_entries_max: int = 1
    # required: bool = True

    def __post_init__(self):
        # Enforce data types
        self.name = str(self.name)
        self.job_input_type = validator.enforce_enum(
            self.job_input_type, CstockJobInputType
        )
        self.web_input_type = validator.enforce_enum(
            self.web_input_type, CstockWebInputType
        )
        # Validate entries are as expected
        self._validate()

    def _validate(self):
        validation_errors = []
        if not self.name:
            validation_errors.append("'name' is empty")

        # Validate min/max values
        if self.num_entries_max > HARD_LIMIT_NUM_ENTRIES_MAX:
            self.num_entries_max = HARD_LIMIT_NUM_ENTRIES_MAX
            logger.warning(
                f"Clamped num_entrie_max ({self.num_entries_max}) to hard limit: {HARD_LIMIT_NUM_ENTRIES_MAX}"
            )
        if self.num_entries_min < 0:
            logger.warning(
                f"Clamped num_entries_min ({self.num_entries_min}) to 0."
            )
            self.num_entries_min = 0
        if self.num_entries_min > self.num_entries_max:
            validation_errors.append(
                f"num_entries_min: {self.num_entries_min} cannot be greater than num_entries_max: {self.num_entries_max}."
            )

        if validation_errors:
            raise cstock_errors.CstockInstantiationError(
                cstock_cls=CstockProjectInput,
                errors=validation_errors,
            )

    def __str__(self):
        msg = f"<CstockProjInput '{self.name}' of type '{self.job_input_type}' (UI:'{self.web_input_type}')>"
        return msg



import logging
import inspect
from typing import Type, TypeVar
from typing import Union
from uuid import UUID
from ceonmedia import errors
from ceonmedia.core.base import CstockBaseEnum
from ceonmedia.core.loader import cstock_from_dict

logger = logging.getLogger(__name__)

import logging
import inspect
from typing import Type, TypeVar
from ceonmedia import errors

T = TypeVar("T")

logger = logging.getLogger(__name__)


E = TypeVar("E", bound="CstockBaseEnum")
T = TypeVar("T")


def enforce_enum(received: CstockBaseEnum, expected_cls: Type[E]) -> E:
    """Cast to the expected enum and return"""
    if isinstance(received, expected_cls):
        return received
    if isinstance(received, CstockBaseEnum):  #  Wrong enum type was provided
        # Prevent other Cstock enum types with matching string values from being
        # silently converted
        raise errors.CstockUnknownTypeError(
            received=received, expected=expected_cls
        )
    return expected_cls(received)


def enforce_uuid(received: UUID) -> UUID:
    """Cast to the expected enum and return"""
    if isinstance(received, UUID):
        return received
    return UUID(received)


def enforce_class_instance(item: Union[dict, Type], cls_to_enforce: Type):
    if isinstance(item, cls_to_enforce):
        return item
    try:
        if isinstance(item, dict):
            # Use the cstock loader from_dict to force errors to be thrown
            # even when instantiating dataclasses on a fastapi server (fastapi overrides
            # the default behaviour when cls(**kwargs), causing invalid kwargs to be ignored
            # even though the same code throws an error in non-fast-api contexts.)
            item = cstock_from_dict(item, cls_to_enforce)
            return item
        if isinstance(item, list):
            item = cls_to_enforce(*item)
            return item
        item = cls_to_enforce(item)
        return item
    except Exception as e:
        raise errors.CstockInstantiationError(
            cstock_cls=cls_to_enforce,
            errors=[
                f"Could not instantiate class instance, received item of type {type(item)} could not be instantiated as an instance of {cls_to_enforce.__name__}: {item=} {e=}"
            ],
        )


def enforce_list_of_instances(list_of_objects, cls_to_enforce: Type):
    if not isinstance(list_of_objects, list):
        raise Exception(
            "Enforce list of instances expects to receive a list (TODO: Auto-wrap non-list as a list?)"
        )
    enforced_list = [
        enforce_class_instance(list_item, cls_to_enforce)
        for list_item in list_of_objects
    ]
    return enforced_list



QUIET = "%(message)s"
DEV = "[%(name)s:%(lineno)d] %(funcName)s(): %(message)s"
DEV_WITH_LEVEL = "%(levelname)-8s[%(name)s:%(lineno)d] %(funcName)s(): %(message)s"
PROD = "[%(asctime)s] %(levelname)-8s[%(name)s:%(lineno)d] %(funcName)s(): %(message)s "



from abc import ABC, abstractmethod

import hou

import ceonmedia.integrations.hou.hparms as hparms
import ceonmedia.integrations.hou.type_parms as tp
import ceonmedia.integrations.hou.hda_parm_namer as parm_namer
import ceonmedia.core.proj_input_type as cprojtype
import ceonmedia.core.proj_input as cprojinput
import ceonmedia.core.job_input as cjobinput
import ceonmedia.utility as helper

if False: # To prevent flake8 from failing on the string type-hints during CI
    from ceonmedia.core.base import CstockBaseEnum


class IHouParmGroup(ABC):
    @abstractmethod
    def parm_templates(self) -> list:
        """returns the parms as a list of parm_templates"""
        pass

    @abstractmethod
    def parm_data(self) -> list:
        """returns the parms as a list of parm_data objects"""
        pass


class TestInputParmGroup(IHouParmGroup):
    def __init__(self, proj_input: cprojinput.CstockProjInput):
        self.proj_input = proj_input
        self.proj_input_type = proj_input.proj_input_type
        self.parm_name = f"test_{proj_input.name}"

    def parm_templates(self) -> list:
        # parmTemplate = self.create_type_picker_dropdown()
        # type_picker = self.create_type_picker_dropdown()
        print("\n\n---Creating parm_templates---")
        parm_templates = [
            hparms.create_parm(parm_data) for parm_data in self.parm_data(debug=True)
        ]
        helper.cprint(parm_templates, f"Created parm_templates:")
        # parms.append(type_picker)
        # parm_templates.append(hparms.create_parm(self.parm_data()))
        return parm_templates

    def parm_data(self, debug=False) -> list:
        input_info_parms = self.input_info_parms()
        value_entry_parms = self.value_entry_parms()
        parm_data = [*input_info_parms, *value_entry_parms]
        if debug:
            helper.cprint(parm_data, "hda_parm_maker created parm_data: ")
        return parm_data

    def value_picker(self) -> list:
        """Takes the proj_input_type
        returns parm data of the appropriate type"""
        class_lookup = {
            cprojtype.CstockProjInputType.AUDIO: tp.TypeParmFile,
            cprojtype.CstockProjInputType.COLOR: tp.TypeParmColor,
            cprojtype.CstockProjInputType.DOC: tp.TypeParmFile,
            cprojtype.CstockProjInputType.DROPDOWN: tp.TypeParmDropdown,
            cprojtype.CstockProjInputType.IMG: tp.TypeParmImg,
            cprojtype.CstockProjInputType.INT: tp.TypeParmInt,
            cprojtype.CstockProjInputType.TEXT: tp.TypeParmString,
        }
        cls_type_parm = class_lookup[self.proj_input_type]
        parm_datas = cls_type_parm(
            self.parm_name, self.proj_input.type_settings
        ).value_picker()
        return parm_datas

    def input_info_parms(self) -> list:
        heading = hparms.create_heading_parm(
            f"{self.parm_name}_heading",
            f"{self.parm_name}_label",
            f"{self.parm_name}_text",
            as_data=True,
        )
        job_input_name = hparms.create_label_parm(
            f"{self.parm_name}_name",
            f"job input name: ",
            f"{self.proj_input.name}",
            tags={"cstock_attrib": "name"},
            as_data=True,
        )
        job_input_type = hparms.create_label_parm(
            f"{self.parm_name}_type",
            f"job input type: ",
            f"{self.proj_input.job_input_type().value}",
            tags={"cstock_attrib": "job_input_type"},
            as_data=True,
        )
        label_description_text = hparms.create_message_parm(
            f"{self.parm_name}_description_text",
            "Description",
            text=self.proj_input.description,
            tags={"cstock_attrib": "description"},
            as_data=True,
        )
        return [heading, job_input_type, job_input_name, label_description_text]

    def value_entry_parms(self) -> list:
        entry_metadata_fldr = {
            "type": hou.FolderParmTemplate,
            "name": f"{self.parm_name}_fldr_information",
            "label": "Metadata",
            "folder_type": hou.folderType.Collapsible,
            "_children": [
                hparms.create_message_parm(
                    f"{self.parm_name}_metadata_help",
                    "",
                    text=(
                        "Available metadata for this input entry.\n"
                        "Note: Some metadata cannot be generated in Houdini, but "
                        "will be available for use in production.\n"
                        "Metadata is not yet implemented"
                    ),
                    as_data=True,
                ),
            ],
        }
        multiparm = {
            **hparms.create_multiparm_template(
                self.parm_name,
                self.parm_name,
                as_data=True,
            ),
            "_children": self.value_picker(),
        }
        sepparm = {
            "type": hou.SeparatorParmTemplate,
            "name": f"sepparm_end_of_{self.parm_name}",
        }
        return [*self.value_picker(), entry_metadata_fldr, sepparm]


class ProjInputParmGroup(IHouParmGroup):
    def parm_templates(self):
        # parmTemplate = self.create_type_picker_dropdown()
        # type_picker = self.create_type_picker_dropdown()
        parm_templates = []
        # parms.append(type_picker)
        parm_templates.append(hparms.create_parm(self.parm_data()))
        # print("proj input editor UI parm_template() returned: ", parms)
        return parm_templates

    def parm_data(self):
        parm_data = {
            **hparms.create_multiparm_template(
                parm_namer.ParmNamer.PROJ_INPUT.BASE_MULTIPARM,
                "Project Inputs",
                as_data=True,
            ),
            "_children": [
                *self.type_picker_parm(),
                *self.common_parms(),
                *self.create_type_settings(),
                {
                    "type": hou.SeparatorParmTemplate,
                    "name": f"sepparm_end_of_type_settings_#",
                },
            ],
        }
        return parm_data

    def type_picker_parm(self):
        # Get dropdown options
        dropdown_options = []
        for input_type in cprojtype.CstockProjInputType:
            dropdown_options.append(
                {"value": input_type.value, "label": input_type.value}
            )
        # Create parm data
        dropdown_parm_data = hparms.create_dropdown_parm(
            "proj_input_#_type",
            "Input Type",
            dropdown_options,
            tags={"cstock_attrib": "proj_input_type"},
            as_data=True,
        )
        return [dropdown_parm_data]

    def common_parms(self):
        # Parms that will be necessary for every proj_input, regardless of type
        # Returns a list of dictionaries containing parm_data
        label_name_parm_data = hparms.create_heading_parm(
            "proj_input_#_label_name",
            "Show Name",
            text=f'`chs("proj_input_#_name")',
            as_data=True,
        )

        common_properties_parm_data = {
            "type": hou.FolderParmTemplate,
            "name": "proj_input_#_fldr_information",
            "label": "Input Information",
            "folder_type": hou.folderType.Collapsible,
            "_children": [
                hparms.create_text_parm(
                    "proj_input_#_name",
                    "Input Name",
                    tags={"cstock_attrib": "name"},
                    default_value="proj_input_#",
                    as_data=True,
                ),
                hparms.create_toggle_parm(
                    "proj_input_#_required",
                    "Required",
                    tags={"cstock_attrib": "required"},
                    default_value=True,
                    as_data=True,
                ),
                hparms.create_text_parm(
                    "proj_input_#_description",
                    "Input Description",
                    tags={"cstock_attrib": "description"},
                    multiline=True,
                    as_data=True,
                ),
            ],
        }
        return [label_name_parm_data, common_properties_parm_data]

    def create_type_settings(self):
        # Parms that are necessary only for each specific input type
        parms = []
        for input_type in cprojtype.CstockProjInputType:
            name = f"proj_input_#_type_settings_{input_type.value}"
            label = f"{input_type.value} settings"
            _children = ProjInputTypeSettings(input_type).parm_templates()
            # TODO show only if current input_type == this input type
            type_settings_parm_data = {
                "type": hou.FolderParmTemplate,
                "name": name,
                "label": label,
                "folder_type": hou.folderType.Collapsible,
                "conditionals": {
                    hou.parmCondType.HideWhen: f"{{ proj_input_#_type != {input_type.value}}}"
                },
                "_children": _children,
            }
            parms.append(type_settings_parm_data)
        return parms


# TODO refactor, move this to type_parms?
class ProjInputTypeSettings:
    # A map of required type_settings and the relevant parm_type to create.
    TYPE_SETTING_PARMS = {
        cprojtype.CstockProjInputType.DROPDOWN: [
            {
                "type_setting": "dropdown_options",
                "parm_function": hparms.create_multiparm_template,
                "parm_kwargs": {
                    "tags": {"cstock_attrib": "type_settings.dropdown_options"},
                    "child_parm_data": [
                        hparms.create_text_parm(
                            name="proj_input_#_type_setting_dropdown_option_#_label",
                            label="Label",
                            tags={"cstock_attrib": "label"},
                            help="The text that will be shown on the dropdown menu",
                            join_with_next=True,
                            as_data=True,
                        ),
                        hparms.create_text_parm(
                            name="proj_input_#_type_setting_dropdown_option_#_value",
                            label="Value",
                            tags={"cstock_attrib": "value"},
                            help="The value received by the software if the user chooses this option",
                            as_data=True,
                        ),
                        {
                            "type": hou.SeparatorParmTemplate,
                            "name": "proj_input_#_type_settings_dropdown_options_#_sepparm",
                        },
                    ],
                },
            }
        ]
    }
    """ Generate parms for specific proj input types """

    def __init__(self, proj_input_type: cprojtype.CstockProjInputType):
        print("Create ProjInputTypeSettings with proj_input_type: ", proj_input_type)
        self.input_type = proj_input_type

    def parm_templates(self):
        parm_templates = []
        # Use a factory-like pattern to build the relevant parm templates
        print("self.input_type: ", self.input_type)
        type_setting_parms = self.TYPE_SETTING_PARMS.get(self.input_type)
        if not type_setting_parms:
            return []
        for type_setting_parm in type_setting_parms:
            type_setting = type_setting_parm["type_setting"]
            parm_func = type_setting_parm["parm_function"]
            type_kwargs = type_setting_parm["parm_kwargs"]

            # setup parm data to create the type_settings folder
            name = f"proj_input_#_type_setting_{self.input_type.value}_{type_setting}"
            label = f"{type_setting}"

            # Append additional kwargs to type_settings folder
            hide_condition = f"{{ proj_input_#_type != {self.input_type.value}}}"
            conditionals = {"conditionals": {hou.parmCondType.HideWhen: hide_condition}}
            type_kwargs = {
                **type_kwargs,
                **conditionals,
            }

            # Append additional kwargs to all children parms inside the folder
            common_child_kwargs = conditionals
            child_parm_data = type_kwargs["child_parm_data"]
            new_child_parm_data = [
                {**child_parm, **common_child_kwargs} for child_parm in child_parm_data
            ]
            type_kwargs["child_parm_data"] = new_child_parm_data

            # Generate the parm_data
            parm_data = parm_func(name, label, **type_kwargs, as_data=True)
            parm_templates.append(parm_data)
        return parm_templates


class JobInputParmGroup(IHouParmGroup):
    def __init__(self, job_input: cjobinput.CstockJobInput):
        self.job_input = job_input

    def parm_template(self):
        parm_templates = []
        parm_templates.append(hparms.create_parm(self.parm_data()))
        return parm_templates

    def parm_data(self):
        parm_data = {
            **hparms.create_multiparm_template(
                parm_namer.ParmNamer.JOB_INPUT.BASE_MULTIPARM,
                "Job Inputs",
                as_data=True,
            ),
            "_children": [
                *self.common_parms(),
            ],
        }
        return parm_data

    def common_parms(self):
        # Parms that will be necessary for every proj_input, regardless of type
        # Returns a list of dictionaries containing parm_data
        input_entry_metadata_parm_data = {
            "type": hou.FolderParmTemplate,
            "name": "job_input_#_fldr_metadata",
            "label": "Input Information",
            "folder_type": hou.folderType.Collapsible,
            "_children": [
                hparms.create_text_parm(
                    "job_input_#_name",
                    "Input Name",
                    tags={"cstock_attrib": "name"},
                    default_value="job_input_#",
                    as_data=True,
                ),
                hparms.create_text_parm(
                    "job_input_#_value",
                    "Input Description",
                    # tags={"cstock_attrib": "value"},
                    multiline=True,
                    as_data=True,
                ),
            ],
        }
        return [input_entry_metadata_parm_data]



# Provides base classes for proper type hinting
from enum import Enum
from abc import ABC

from ceonmedia.errors import CstockUnknownTypeError


class CstockBaseClass(ABC):
    """A base class for all custom Cstock classes to inherit from."""


class CstockBaseEnum(str, Enum):
    """A base class for all custom Cstock types (enums) to inherit from."""

    # TODO test: does change/affect behaviour if the enum type is passed directly a search value?
    # E.g. CstockProjInputType(CstockProjInputType.IMG) instead of passing a string?
    @classmethod
    def _missing_(cls, value: str):
        # Allow this Enum to match string values case-insensitively
        if isinstance(value, str):
            for member in cls:
                if member.value.lower() == value.lower():
                    return member
        # If not found, raise a CstockUnknownTypeError instead of the default ValueError
        raise CstockUnknownTypeError(received=value, expected=cls)



import json
import logging
from uuid import UUID
from typing import Any

from ceonmedia.core.base import CstockBaseClass

logger = logging.getLogger(__name__)


# Allow handling of non-serializable types
class CstockJsonEncoder(json.JSONEncoder):
    def default(self, value: Any):
        """JSON serialization conversion function."""

        # Handle serialization of otherwise unserializable types
        if isinstance(value, UUID):
            return str(value)

        if isinstance(value, CstockBaseClass):
            return value.__dict__
            # return json.loads(json.dumps(value.__dict__, **DUMP_ARGS))

        # Here you can have other handling for your
        # IPV4, or datetimes, or whatever else you
        # have.

        # Otherwise, default to super
        return super(CstockJsonEncoder, self).default(value)


DUMP_ARGS = {"cls": CstockJsonEncoder, "indent": 4, "sort_keys": True}


# TODO allow kwargs in case user want to overwrite defaults
def dumps(some_dict: dict):
    return json.dumps(some_dict, **DUMP_ARGS)


# TODO allow kwargs in case user want to overwrite defaults
def dump(some_dict, fp):
    return json.dump(some_dict, fp, **DUMP_ARGS)



import hou
from abc import ABC, abstractmethod
from typing import List

import ceonmedia.integrations.hou.hparms as hparms
import ceonmedia.integrations.hou.parm_processor as pp
import ceonmedia.integrations.hou.hda_parm_maker as parm_maker
import ceonmedia.integrations.hou.hda_parm_namer as parm_namer
import ceonmedia.core.proj_input as cprojinput
import ceonmedia.core.job_input as cjobinput
import ceonmedia.integrations.hou.exceptions as hou_err
import ceonmedia.utility as utility


"""
class MappedParmGroup:
    # TODO move this to multiparm processor? Mostly only necessary when dealing with
    # multiple multiparm instances?

    def __init__(self, object_instance, parm_group):
        # TODO test that passed object instances will correctly updated by reference
        self.object_instance = object_instance
        self.parm_group = parm_group
        self.visible_parms = self.get_visible_parms()
        self.multiparm_handler = mpp.MultiparmProcessorDictKeyFromParmTag(
            "cstock_dict_key"
        )

    def get_visible_parms(self):
        self.parm_group[0].node().updateParmStates()
        visible_parms = [
            parm for parm in self.parm_group if not hparms.is_parm_hidden(parm)
        ]
        self.visible_parms = visible_parms
        return visible_parms

    def set_parms(self):
        for parm in self.parm_group:
            # TODO better way to handle the missing parm problem?
            # From a higher level, first set multiparm instances, then set the values?
            try:
                tags = parm.parmTemplate().tags()
                attrib_name = tags.get("cstock_attrib")
                if not attrib_name:
                    return
                new_value = self.object_instance.to_dict().get(attrib_name)
                if not new_value:
                    print(
                        f"WARNING: Skipping parm({parm.name()}), attrib({attrib_name}) not found in class instance {self.object_instance.__class__}."
                    )
                    return
                print(
                    f"Setting parm({parm.name()}) from attrib({attrib_name}) to new_value: {new_value}"
                )
                hparms.set_parm(
                    parm, new_value, multiparm_processor=self.multiparm_handler
                )
            except hou.ObjectWasDeleted:
                print("WARNING: Skipped deleted parm")

    def from_parms(self):
        print("")
        proj_input_dict = {}
        for parm in self.visible_parms:
            tags = parm.parmTemplate().tags()
            attrib_name = tags.get("cstock_attrib")
            if attrib_name:
                new_value = hparms.read_parm(
                    parm, multiparm_processor=self.multiparm_handler
                )
                proj_input_dict[attrib_name] = new_value
                # print(f"got parm({parm.name()}) value for attrib({attrib_name}): {new_value}")
        # Create object instance
        processed_proj_input_dict = utility.dict_dotnotation_to_nested(proj_input_dict)
        utility.print_dict(
            processed_proj_input_dict, msg="Created processed_proj_input_dict: "
        )
        new_object_instance = self.object_instance.from_dict(processed_proj_input_dict)
        return new_object_instance
"""

"""
def create_mapped_parm_groups(multiparm, object_instances):
    # TODO split into two functions:
    # one for creating new class instances from existing parms.
    # One for creating new parms from existing instances.
    parm_instance_groups = hparms.get_multiparm_parms_grouped_by_instance(multiparm)
    mapped_parm_groups = []

    # If object_instances is not a list, assume a class definition was received
    # Map the parms to the desired class definition instead of an instance
    create_new_instances = False if isinstance(object_instances, list) else True

    for instance_index, parm_group in parm_instance_groups.items():
        instance_to_map = (
            object_instances
            if create_new_instances
            else object_instances[int(instance_index) - 1]
        )
        mapped_parm_group = MappedParmGroup(instance_to_map, parm_group)
        mapped_parm_groups.append(mapped_parm_group)
    return mapped_parm_groups
"""


class IUIManager(ABC):
    @abstractmethod
    def to_parms():
        """Loads the python classes into the houdini UI"""

    @abstractmethod
    def from_parms():
        """Creates new classe by reading the current houdini UI."""


class ProjInputUIManager(IUIManager):
    def __init__(self, node):
        self.node = node
        self.multiparm_name = parm_namer.ParmNamer.PROJ_INPUT.BASE_MULTIPARM
        self.parms_processor = pp.ParmsProcessor(
            pp.ParmProcessorFromParmTag("cstock_attrib")
        )

    def load_proj_input_editor(self):
        proj_input_editor_parms = parm_maker.ProjInputParmGroup().parm_templates()
        print("proj_input_editor_parms: ", proj_input_editor_parms)
        hparms.set_folder_parms(
            self.node,
            parm_namer.ParmNamer.PROJ_INPUT.TARGET_FOLDER,
            proj_input_editor_parms,
        )

    def to_parms(self, proj_inputs: List[cprojinput.CstockProjInput]):
        # Create multiparm parameter instances
        multiparm = self.get_multiparm(self.multiparm_name)
        num_inputs = len(proj_inputs)
        multiparm.set(num_inputs)

        # Map each proj_input object with the corresponding group of parameters
        parm_instance_groups = hparms.get_multiparm_parms_grouped_by_instance(multiparm)
        for index, parms in parm_instance_groups.items():
            proj_input = proj_inputs[int(index) - 1]
            print("Setting parms for proj_input: ", proj_input)
            values_to_set = proj_input.to_dict()
            utility.cprint(values_to_set, "values_to_set: ")
            self.parms_processor.set_values(parms, values_to_set)

    def from_parms(self) -> List[cprojinput.CstockProjInput]:
        """
        Creates proj inputs from current parm settings.
        Returns the new proj_input instances
        """
        multiparm = self.get_multiparm(self.multiparm_name)
        new_proj_inputs = []

        print("")
        parm_instance_groups = hparms.get_multiparm_parms_grouped_by_instance(multiparm)
        for _, parms in parm_instance_groups.items():
            # Filter to remove parms that are hidden from the UI
            filtered_parms = hparms.get_visible_parms(parms)
            values_from_parms = self.parms_processor.get_values(filtered_parms)
            new_proj_input = cprojinput.CstockProjInput.from_dict(values_from_parms)
            new_proj_inputs.append(new_proj_input)
        utility.cprint(new_proj_inputs, msg="proj_inputs after getting from parms: ")
        return new_proj_inputs

    def get_multiparm(self, multiparm_name):
        """Helper function for validating expected multiparms"""
        multiparm = self.node.parm(multiparm_name)
        if not multiparm:
            raise hou_err.MissingParmException(multiparm_name, self.node.path())
        if not multiparm.isMultiParmParent():
            raise Exception(f"Parm({multiparm.name()}) is not a multiparm")
        return multiparm


class TestInputUIManager(IUIManager):
    def __init__(self, node):
        self.node = node
        self.target_folder_name = parm_namer.ParmNamer.TEST_INPUT.TARGET_FOLDER
        self.parms_processor = pp.ParmsProcessor(
            pp.ParmProcessorFromParmTag("cstock_attrib")
        )

    def from_parms(self) -> List[cjobinput.CstockJobInput]:
        target_folder_parm_template = self.node.parmTemplateGroup().findFolder(
            self.target_folder_name
        )
        test_input_folders = hparms.parm_templates_in_folder(
            target_folder_parm_template
        )
        print("Got test_input_folders: ", test_input_folders)
        new_job_inputs = []
        for proj_input_folder in test_input_folders:
            print("Got folder parm: ", proj_input_folder)
            parm_tuples_in_folder = hparms.parm_tuples_in_folder(
                self.node, proj_input_folder
            )
            utility.cprint(parm_tuples_in_folder, "parm_tuples_in_folder: ")

            values_from_parms = self.parms_processor.get_values(parm_tuples_in_folder)
            print("\tCreated values: ", values_from_parms)

            # Make sure that 'values' is a list
            if not isinstance(values_from_parms["values"], list):
                values_from_parms["values"] = [values_from_parms["values"]]

            new_job_input = cjobinput.CstockJobInput.from_dict(values_from_parms)
            print("\tCreated job_input: ", new_job_input)
            new_job_inputs.append(new_job_input)

        utility.cprint(new_job_inputs, msg="job_inputs created from test parms: ")
        return new_job_inputs

    def to_parms(self, proj_inputs: List[cprojinput.CstockProjInput]):
        """
        For each proj_input, create a multiparm.
        Each multiparm instance corresponds to one user entry for this input, and
        includes associated metadata.
        The whole multiparm template will be created according to the job_input type.
        """
        new_parms = []
        for proj_input in proj_inputs:
            fldr_name = f"test_{proj_input.name}_fldr"
            folder_parms = parm_maker.TestInputParmGroup(proj_input).parm_data()
            new_folder = hparms.create_folder_parm(
                fldr_name,
                fldr_name,
                child_parm_data=folder_parms,
                folder_config="collapsible",
                as_data=True,
            )
            new_parm = hparms.create_parm(new_folder)
            new_parms.append(new_parm)
        # Flatten the list
        # new_parms_flatlist = [
        # new_parm for parm_list in new_parms for new_parm in parm_list
        # ]
        utility.cprint(new_parms, "hda_ui_manager to_parms: setting new parms: ")
        hparms.set_folder_parms(
            self.node,
            parm_namer.ParmNamer.TEST_INPUT.TARGET_FOLDER,
            new_parms,
        )


class JobInputUIManager(IUIManager):
    def __init__(self, node, job_inputs=[]):
        self.node = node
        self.job_inputs = job_inputs

    def to_parms(self):
        """
        For each job_input, create a multiparm.
        Each multiparm instance corresponds to one user entry for this input, and
        contains associated metadata.
        The whole multiparm template will be created according to the job_input type.
        """
        imported_job_input_parms = parm_maker.HDAParmMaker().job_input_parms()
        hparms.set_folder_parms(
            self.node,
            parm_namer.ParmNamer.JOB_INPUT.IMPORT_FOLDER,
            imported_job_input_parms,
        )

    def from_parms(self):
        pass

    def from_proj_inputs(self, list_of_proj_inputs):
        new_job_inputs = [
            cjobinput.CstockJobInput.from_proj_input(proj_input)
            for proj_input in list_of_proj_inputs
        ]
        self.job_inputs = new_job_inputs
        return new_job_inputs


class HDAUIManager:
    def __init__(self, node, job_inputs=[]):
        print("Init HDAUIManager! node: ", node)
        # TODO refactor: the UI managers should not store inputs, just provide logic
        # to set and fetch from/to parms.
        # Only the HDAUIManager handles the proj/job inputs
        self.node = node
        self.proj_inputs = ProjInputUIManager(node)
        self.test_inputs = TestInputUIManager(node)
        self.job_inputs = JobInputUIManager(node, job_inputs)
        # TODO pipeline

    def get_proj_inputs(self):
        return self.proj_inputs.from_parms()

    def import_test_inputs(self):
        # self.proj_inputs.from_parms()
        proj_inputs = self.proj_inputs.from_parms()
        self.test_inputs.to_parms(proj_inputs)
        print("test inputs to parms:")
        for proj_input in proj_inputs:
            print("test_input: ", proj_input)

    # def export_test_inputs(self):
    # job_inputs = self.job_inputs.from_parms()
    # print("TODO export job_inputs: ")
    # for job_input in job_inputs:
    # print(job_input)



# Import/Export for a ceonmedia job
# To avoid polluting the base classes implement (de)serialization tools in this module directly,
import json
from dataclasses import asdict
from ceonmedia.core import CstockJob
from ceonmedia.core.job_input import CstockJobInput, CstockJobInputType

# POSSIBLE ALTERNATIVE dataclass-wizard.
#    https://dataclass-wizard.readthedocs.io/en/latest/
# dataclass_wizard:
# - Expects enums to use VALUES by default: 'color' not 'COLOR'
# - Only provider one error at a time on failure.
# - Doesn't require schema class, just use fromdict and asdict methods directly.
# - Good documentation and seems to be easy to config

# from dataclass_wizard import fromdict, asdict, DumpMeta


# marshmallow:
# - Gives cleaerer errors (reports all together)
# - Expects enums to use NAME not VALUE, e.g. 'COLOR' not 'color'
# - Requires instantiating the schema class first

# Chose to use marshmallow for the multi-error reporting. NOTE: enum NAME vs VALUE could cause
# some compatibility problems with workflows that expect values? It's possible to configures
# marshmallow but haven't looked into it yet and seems to not support enums contained
# within a list?

# Keep dataclass_wizard available as a backup in case the enum naming convention causes problems in the
# future? (Or learn how to configure marshmallow to use enum values)

# import marshmallow_dataclass

# Create a marshmallow schema instance to use for (de)serialization operations.
# job_schema = marshmallow_dataclass.class_schema(CstockJob)()


def _job_input_from_dict(dict_data) -> CstockJobInput:
    prepared_dict = dict_data.copy()
    try:
        # Load from enum VALUE
        job_input_type = CstockJobInputType(dict_data["job_input_type"])
    except ValueError:
        # Load from enum NAME
        job_input_type = CstockJobInputType[dict_data["job_input_type"]]
    prepared_dict["job_input_type"] = job_input_type
    return CstockJobInput(**prepared_dict)


def from_dict(dict_data) -> CstockJob:
    """
    Load a CstockJob instance from a received json dict.
    """
    prepared_dict = dict_data.copy()
    prepared_dict["job_inputs"] = [
        _job_input_from_dict(data) for data in dict_data["job_inputs"]
    ]
    return CstockJob(**prepared_dict)


def from_file(file_path: str) -> CstockJob:
    with open(file_path, "r") as f:
        data = json.load(f)
    job = from_dict(data)
    return job


def to_file(file_path: str, job: CstockJob) -> str:
    """Save a CstockJob to a json file.
    Returns: The path of the newly saved file as a string"""

    job_dict = to_dict(job)

    with open(file_path, "w") as f:
        json.dump(job_dict, f, indent=2)
    return file_path


def to_dict(job: CstockJob) -> dict:
    """Save a CstockJob to a json file.
    Returns: The path of the newly saved file as a string"""

    job_dict = asdict(job)

    return job_dict



import json
import logging
from typing import Union
from pathlib import Path
from uuid import UUID
from dataclasses import asdict

from ceonmedia.core.base import CstockBaseClass
from ceonmedia.core.base import CstockBaseEnum

logger = logging.getLogger(__name__)


class CstockJsonEncoder(json.JSONEncoder):
    """Allow JSON saving of non-serializable types"""

    def default(self, obj):
        if isinstance(obj, UUID):
            # if the obj is uuid, we simply return the value of uuid
            return str(obj)

        if isinstance(obj, CstockBaseClass):
            return asdict(obj)

        return json.JSONEncoder.default(self, obj)


DUMP_ARGS = {"cls": CstockJsonEncoder, "indent": 4, "sort_keys": True}


def dumps(some_dict: dict):
    return json.dumps(some_dict, **DUMP_ARGS)


def dump(some_dict: dict, fp):
    return json.dump(some_dict, fp, **DUMP_ARGS)


def save_json(json_filepath: Union[str, Path], json_data: dict):
    """
    Save json file.
    Uses a custom Encoder to handle unserializable types.
    """
    logger.debug(f"Saving json data: {json_data}")
    file_dir = Path(json_filepath).parent
    if not file_dir.is_dir():
        logger.info(f"Creating dir: {file_dir}")
        file_dir.mkdir(exist_ok=True, parents=True)
    with open(json_filepath, "w") as f:
        f.write(dumps(json_data))
    logger.info(f"Saved json file: {json_filepath}")


def save_ceonmedia_json(
    json_filepath: Union[str, Path], ceonmedia_instance: CstockBaseClass
):
    """
    Save a ceonmedia_instance to a json file.
    Uses a custom Encoder to handle unserializable types.
    """
    logger.info(f"Saving json from ceonmedia instance: {ceonmedia_instance}")
    dict_data = asdict(ceonmedia_instance)
    logger.debug(f"Created dict data: {dict_data}")
    with open(json_filepath, "w") as f:
        f.write(dumps(dict_data))
    logger.info(f"Saved json file: {json_filepath}")



from uuid import UUID, uuid4
from ceonmedia.core.render import CstockRenderPipeline, CstockRenderAppType
from . import json_tests
from ceonmedia.tests.dicts import render_pipeline_dicts


TEST_CLS = CstockRenderPipeline
TEST_DICTS = render_pipeline_dicts


def test_render_pipeline_json_serialization() -> None:
    for test_dict in TEST_DICTS.SUCCESS_CASES:
        test_object = TEST_CLS(**test_dict)
        json_tests.run_json_serialization_tests(cstock_object=test_object)


# TODO setup proper testing/assertion of failure cases
# def test_render_pipeline_json_serialization_fail() -> None:
#     for test_dict in TEST_DICTS.FAIL_CASES:
#         test_object = TEST_CLS(**test_dict)
#         json_tests.run_json_serialization_tests(cstock_object=test_object)



import logging
from typing import Union
from pathlib import Path
from json.decoder import JSONDecodeError

from . import exporter
from . import importer
from . import errors

logger = logging.getLogger(__name__)


def write_json_file(dict_data: dict, json_file_path: Union[Path, str]) -> str:
    json_data = exporter.dumps(dict_data)
    ensure_dir(Path(json_file_path).parent)
    with open(json_file_path, "w") as f:
        f.write(json_data)
    return json_data


def read_json_file(
    json_filepath: Union[str, Path], raise_empty_file=True
) -> str:
    # read json file
    if not Path(json_filepath).is_file():
        raise FileNotFoundError(
            f"Unable to find json file for json_file_path: {json_filepath}"
        )
    json_data = {}
    # Get data from file
    try:
        with open(str(json_filepath)) as f:
            json_data = importer.load(f)
    except JSONDecodeError as e:
        raise errors.InvalidJSONFileError(
            json_filepath, errors=["Failed to decode file", str(e)]
        )

    if not len(json_data) > 0:  # No data in file
        # print("No data found in file")
        if raise_empty_file:
            raise errors.InvalidJSONFileError(
                json_filepath,
                errors=["File does not contain data (empty JSON file)"],
            )
    # print(f"Got data in JSON({json_filepath}): {json_data}")

    return json_data


def ensure_dir(filepath: Union[Path, str]):
    """Create the dir if it doesn't already exist"""
    filepath = Path(filepath)
    if not filepath.exists():
        logger.info(f"Created new dir for json export: {filepath}")
        filepath.mkdir(parents=True)



#
# Master interface for functions which are called from within the houdini HDA
#
# Handles the import/export of parms.
# Determines from where to read the in/out files.
# Handles node parm organisation with folders/groups.
# Does not handle parm generataion. Parm generation is handled by the parm_maker.
#
from pprint import pprint
from typing import List
import importlib

import hou
from ceonmedia import CstockJobInput
from ceonmedia import CstockJob

# from ceonmedia.log import printify
# from ceonmedia.json_io import cstock_from_json
from ceonmedia import json_io

from . import parm_maker

importlib.reload(parm_maker)
importlib.reload(parm_maker.hparms)


class HoudiniHDA:
    """
    Interface for interacting with the houdini HDA node.
    This is the only thing that should need to be imported into Hou.
    -
    Note: For development, it may be necessary to reload submodules inside of houdini with
    importlib.reload(module)
    -
    """

    def __init__(self, node: "hou.Node"):
        self.node = node
        self.job_inputs_import_fldr = ("Imports",)

    def _recook_scene(self):
        print("Recooking Scene ..")
        for node in parm_maker.hparms.all_nodes():
            print(f"\tCooking {node.path()}")
            try:
                node.cook(force=True)
            except Exception:
                print(f"WARNING: Failed to cook node: {node}")

    def _imported_parms(self):
        """Return all parms in the target import folder"""
        imported_parms = self.node.parmsInFolder(self.job_inputs_import_fldr)
        print(
            f"Got import parms from folder '{self.job_inputs_import_fldr}': "
        )
        for parm in imported_parms:
            print(f"\t{parm.name()}: {parm.eval()}")
        if not imported_parms:
            print("\t(None Found)")
        return imported_parms

    def _set_imported_parm_values_to_default(self):
        print(f"Setting parm values to new defaults...")
        imported_node_parms = self.node.parmsInFolder(
            self.job_inputs_import_fldr
        )
        for parm in imported_node_parms:
            parm.revertToDefaults()

    def _apply_parms(self, parms: List[hou.FolderParmTemplate]):
        """Handles applying/updating the given parms to the node."""
        # just to print the parms for testing
        self._imported_parms()

        print("Received list of FolderPartmTemplates to apply:")
        for parm in parms:
            print(f"\t{parm}")

        print(f"Applying {len(parms)} folderParmTemplates to node ...")
        parm_maker.hparms.set_folder_parms(
            self.node, self.job_inputs_import_fldr, parms
        )
        print("Applied parms!")

        # Just printing for debugging
        self._imported_parms()

        # Just to print for debugging.
        parm_maker.hparms.get_dependant_nodes(self.node)

    def import_job_inputs(self, ceonmedia_job_json_filepath: str):
        """Import job_inputs from a file and display them as locked nodes.
        inputs:
            node: Assumed to be the ceonmedia HDA which is calling this function.
            ceonmedia_job_json_filepath: The filepath to a valid ceonmedia_job.json file which includes
        'job_inputs' as a top level key"""
        print(f"Got job_inputs_json_file: {ceonmedia_job_json_filepath}")
        print("Reading json file ...")
        cstock_job = json_io.job.from_file(ceonmedia_job_json_filepath)
        job_inputs = cstock_job.job_inputs
        print("Got job_inputs:")
        pprint(job_inputs)
        folders_with_parms = []
        print("")
        print("Creating job input parms ... ")
        for job_input in job_inputs:
            job_input_parms = parm_maker.job_input_parms(job_input)
            print("Created new job_input_parms: ")
            for job_input_parm in job_input_parms:
                print(f"\t{job_input_parm}")

            # Wrap each individual job_inut in a folder
            folder_parm = parm_maker.hparms.create_folder_parm(
                job_input.name, job_input.name
            )
            folder_parm.setParmTemplates(job_input_parms)
            folder_parm.setFolderType(hou.folderType.Simple)
            # folder_parm.setFolderType(hou.folderType.Collapsible)
            folders_with_parms.append(folder_parm)

            # Apply and disable
            # _apply_parms(node, [folder_parm])
            # parm_maker.hparms.disable_parms(node, job_input_parms)
        print("")
        self._apply_parms(folders_with_parms)
        # Applying parms will NOT change the values if the parm already exists.
        # Therefore we need to set the values to match the new defaults.
        self._set_imported_parm_values_to_default()
        # For some reason when disabling parms, the UI doesn't update correctly on re-import.
        # Even when un-disabling before applying
        # parm_maker.hparms.disable_parms(node, folders_with_parms, disabled=False)
        # _apply_parms(node, folders_with_parms)
        # parm_maker.hparms.disable_parms(node, folders_with_parms)



import logging
from pathlib import Path

from ceonmedia import CstockProject
from ceonmedia import CstockJobInputType
from ceonmedia import CstockProjectInput
from ceonmedia import CstockRenderPipeline
from ceonmedia import CstockRenderTask
from ceonmedia import CstockRenderAppType
from ceonmedia.core import CstockFileReference
from ceonmedia.core.render_task import CstockRenderTaskAppSettingsHou
from ceonmedia.job_maker import value_picker
from ceonmedia.job_maker import job_maker

logger = logging.getLogger(__name__)

JOB_INPUT_FILES_DIR = "/mnt/FileStorage/Dayne/Web/proj_ceonmedia/assets/project_input_files/job_test_rendering_files"
JOBS_DIR = Path("./temp_testfiles_created_jobs").resolve()

# Required to instantiate CstockProject but not relevant in this testcase
RENDER_TASK = CstockRenderTask(
    task_name="hou_render", 
    app_type=CstockRenderAppType.HOU, 
    app_version="20.0", 
    app_render_settings=CstockRenderTaskAppSettingsHou(target_node="/out/karm1", frames="0 25"),
    task_input=CstockFileReference("path/to/hipfile.hiplc"),
    task_output="renderoutput.$F4.png"
)
RENDER_PIPELINE = CstockRenderPipeline(render_tasks=[RENDER_TASK], output_task="hou_render")

PROJECT_INPUTS = {
    "img": CstockProjectInput(
        name="single_image",
        job_input_type=CstockJobInputType.IMG,
        web_input_type=CstockProjectInput.web_input_types.IMG,
        description="A single image TODO system to enforce limit on num inputs",
    ),
    "doc": CstockProjectInput(
        name="image_from_doc",
        job_input_type=CstockJobInputType.IMG,
        web_input_type=CstockProjectInput.web_input_types.DOC,
        description="A single image TODO system to enforce limit on num inputs",
    ),
    "col": CstockProjectInput(
        name="single_color",
        job_input_type=CstockJobInputType.COLOR,
        web_input_type=CstockProjectInput.web_input_types.COLOR,
        description="A single color",
    ),
    "dropdown": CstockProjectInput(
        name="dropdown_string",
        job_input_type=CstockJobInputType.STRING,
        web_input_type=CstockProjectInput.web_input_types.DROPDOWN,
        type_settings={
            "dropdown_options": [
                {
                    "value": "dropdown_option_value_0",
                    "label": "Option Label 0",
                },
                {
                    "value": "dropdown_option_value_1",
                    "label": "Option Label 1",
                },
                {
                    "value": "dropdown_option_value_2",
                    "label": "Option Label 2",
                },
            ]
        },
        description="A single image TODO system to enforce limit on num inputs",
    ),
}

# TODO load projects form a set of dirs / json files?
PROJECTS_TO_TEST = [
    CstockProject(
        project_inputs=[PROJECT_INPUTS["img"], PROJECT_INPUTS["dropdown"]],
        render_pipeline=RENDER_PIPELINE,
    ),
    CstockProject(
        project_inputs=[PROJECT_INPUTS["doc"], PROJECT_INPUTS["col"]],
        render_pipeline=RENDER_PIPELINE,
    ),
]


def test_create_job_random_inputs():
    project = PROJECTS_TO_TEST[0]
    vp = value_picker.JobInputValuePickerFromFilesRandom(JOB_INPUT_FILES_DIR)
    for project in PROJECTS_TO_TEST:
        job = job_maker.create_job_for_project(project, vp, JOBS_DIR)
        logger.info("Created job: ")
        logger.info(f"{job}")
        logger.warning("TODO: Assert job is valid")


def test_create_job_ordered_inputs():
    project = PROJECTS_TO_TEST[0]
    vp = value_picker.JobInputValuePickerFromFilesOrdered(JOB_INPUT_FILES_DIR)
    for project in PROJECTS_TO_TEST:
        job = job_maker.create_job_for_project(project, vp, JOBS_DIR)
        logger.info("Created job: ")
        logger.info(f"{job}")
        logger.warning("TODO: Assert job is valid")
